# -*- coding: utf-8 -*-
import pandas as pd
import pymatgen.core
# @Time  : 2023/1/31 16:55
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License
from ase import Atoms
# from ase.optimize import BFGS, GPMin
from ase.calculators.emt import EMT
import numpy as np
from hpc2ml.data.io import aaa
from pymatgen.core import Structure

from hpc2ml.data.structuretodata import StructureToData, PAddXArray, PAddPBCEdgeDistance

from hpc2ml.nn.cggru import CGGRU

from hpc2ml.optimize.calculator import GNNEICalculater
from hpc2ml.optimize.simpleeimin import SimpleEIGPMin
from hpc2ml.optimize.regpmin import ResampleGPMin
from hpc2ml.optimize.eimin import EIMin
from hpc2ml.optimize.strengthengpmin import StrengthenGPMin
from hpc2ml.optimize.basegpmin import BaseGPMin

d = 0.9575
t = np.pi / 180 * 104.51
water = Atoms('H2O',
              positions=[(d, 0, 0),
                         (d * np.cos(t), d * np.sin(t), 0),
                         (0, 0, 0)],
              calculator=EMT())

dyn = BaseGPMin(water)
# dyn = ResampleGPMin(water)
# dyn = StrengthenGPMin(water,update_hyperparams=True, batch_size=10)
# dyn = EIMin(water)
# dyn = SimpleEIGPMin(water)
dyn.run(fmax=0.01)


"""This part contains some general support function to deal with data.
such as one function could be used in both preprocessing and transforms
Make sure the original source in function. (the files are not import from other code in this package.)
"""



import inspect
import logging
from typing import Optional, Union, List

from .config import Config


def instantiate_from_cls_name(
        module,
        class_name: str,
        prefix: Optional[Union[str, List[str]]] = [],
        positional_args: dict = {},
        optional_args: Optional[dict] = None,
        all_args: Optional[dict] = None,
        remove_kwargs: bool = True,
        return_args_only: bool = False,
):
    """Initialize a class based on a string class name

    Args:
    module: the module to import the class, i.e. torch.optim
    class_name: the string name of the class, i.e. "CosineAnnealingWarmRestarts"
    positional_args (dict): positional arguments
    optional_args (optional, dict): optional arguments
    all_args (dict): list of all candidate parameters tha could potentially match the argument list
    remove_kwargs: if True, ignore the kwargs argument in the init funciton
        same definition as the one in Config.from_function
    return_args_only (bool): if True, do not instantiate, only return the arguments

    Returns:

    instance: the instance
    optional_args (dict):
    """

    if class_name is None:
        raise NameError("class_name type is not defined ")

    # first obtain a list of all classes in this module
    class_list = inspect.getmembers(module, inspect.isclass)
    class_dict = {}
    for k, v in class_list:
        class_dict[k] = v

    # find the matching class
    the_class = class_dict.get(class_name, None)
    if the_class is None:
        raise NameError(f"{class_name} type is not found in {module.__name__} module")

    return instantiate(
        builder=the_class,
        prefix=prefix,
        positional_args=positional_args,
        optional_args=optional_args,
        all_args=all_args,
        remove_kwargs=remove_kwargs,
        return_args_only=return_args_only,
    )


def instantiate(
        builder,
        prefix: Optional[Union[str, List[str]]] = [],
        positional_args: dict = {},
        optional_args: dict = None,
        all_args: dict = None,
        remove_kwargs: bool = True,
        return_args_only: bool = False,
        parent_builders: list = [],
):
    """Automatic initializing class instance by matching keys in the parameter dictionary to the constructor function.

    Keys that are exactly the same, or with a 'prefix_' in all_args, optional_args will be used.
    Priority:

        all_args[key] < all_args[prefix_key] < optional_args[key] < optional_args[prefix_key] < positional_args

    Args:
        builder: the type of the instance
        prefix: the prefix used to address the parameter keys
        positional_args: the arguments used for input. These arguments have the top priority.
        optional_args: the second priority group to search for keys.
        all_args: the third priority group to search for keys.
        remove_kwargs: if True, ignore the kwargs argument in the init funciton
            same definition as the one in Config.from_function
        return_args_only (bool): if True, do not instantiate, only return the arguments
    """

    prefix_list = [builder.__name__] if inspect.isclass(builder) else []
    if isinstance(prefix, str):
        prefix_list += [prefix]
    elif isinstance(prefix, list):
        prefix_list += prefix
    else:
        raise ValueError(f"prefix has the wrong type {type(prefix)}")

    # detect the input parameters needed from params
    config = Config.from_class(builder, remove_kwargs=remove_kwargs)

    # be strict about _kwargs keys:
    allow = config.allow_list()
    for key in allow:
        bname = key[:-7]
        if key.endswith("_kwargs") and bname not in allow:
            raise KeyError(
                f"Instantiating {builder.__name__}: found kwargs argument `{key}`, but no parameter `{bname}` for the corresponding builder. (Did you rename `{bname}` but forget to change `{bname}_kwargs`?) Either add a parameter for `{bname}` if you are trying to allow construction of a submodule, or, if `{bname}_kwargs` is just supposed to be a dictionary, rename it without `_kwargs`."
            )
    del allow

    key_mapping = {}
    if all_args is not None:
        # fetch paratemeters that directly match the name
        _keys = config.update(all_args)
        key_mapping["all"] = {k: k for k in _keys}
        # fetch paratemeters that match prefix + "_" + name
        for idx, prefix_str in enumerate(prefix_list):
            _keys = config.update_w_prefix(
                all_args,
                prefix=prefix_str,
            )
            key_mapping["all"].update(_keys)

    if optional_args is not None:
        # fetch paratemeters that directly match the name
        _keys = config.update(optional_args)
        key_mapping["optional"] = {k: k for k in _keys}
        # fetch paratemeters that match prefix + "_" + name
        for idx, prefix_str in enumerate(prefix_list):
            _keys = config.update_w_prefix(
                optional_args,
                prefix=prefix_str,
            )
            key_mapping["optional"].update(_keys)

    # for logging only, remove the overlapped keys
    if "all" in key_mapping and "optional" in key_mapping:
        key_mapping["all"] = {
            k: v
            for k, v in key_mapping["all"].items()
            if k not in key_mapping["optional"]
        }

    final_optional_args = dict(config)

    # for nested argument, it is possible that the positional args contain unnecesary keys
    if len(parent_builders) > 0:
        _positional_args = {
            k: v for k, v in positional_args.items() if k in config.allow_list()
        }
        positional_args = _positional_args

    init_args = final_optional_args.copy()
    init_args.update(positional_args)

    # find out argument for the nested keyword
    search_keys = [key for key in init_args if key + "_kwargs" in config.allow_list()]
    for key in search_keys:
        sub_builder = init_args[key]
        if sub_builder is None:
            # if the builder is None, skip it
            continue

        if not (callable(sub_builder) or inspect.isclass(sub_builder)):
            raise ValueError(
                f"Builder for submodule `{key}` must be a callable or a class, got `{sub_builder!r}` instead."
            )

        # add double check to avoid cycle
        # only overwrite the optional argument, not the positional ones
        if (
                sub_builder not in parent_builders
                and key + "_kwargs" not in positional_args
        ):
            sub_prefix_list = [sub_builder.__name__, key]
            for prefix in prefix_list:
                sub_prefix_list = sub_prefix_list + [
                    prefix,
                    prefix + "_" + key,
                ]

            nested_km, nested_kwargs = instantiate(
                sub_builder,
                prefix=sub_prefix_list,
                positional_args=positional_args,
                optional_args=optional_args,
                all_args=all_args,
                remove_kwargs=remove_kwargs,
                return_args_only=True,
                parent_builders=[builder] + parent_builders,
            )
            # the values in kwargs get higher priority
            nested_kwargs.update(final_optional_args.get(key + "_kwargs", {}))
            final_optional_args[key + "_kwargs"] = nested_kwargs

            for t in key_mapping:
                key_mapping[t].update(
                    {key + "_kwargs." + k: v for k, v in nested_km[t].items()}
                )
        elif sub_builder in parent_builders:
            raise RuntimeError(
                f"cyclic recursion in builder {parent_builders} {sub_builder}"
            )
        elif not callable(sub_builder) and not inspect.isclass(sub_builder):
            logging.warning(f"subbuilder is not callable {sub_builder}")
        elif key + "_kwargs" in positional_args:
            logging.warning(
                f"skip searching for nested argument because {key}_kwargs are defined in positional arguments"
            )

    # remove duplicates
    for key in positional_args:
        final_optional_args.pop(key, None)
        for t in key_mapping:
            key_mapping[t].pop(key, None)

    if return_args_only:
        return key_mapping, final_optional_args

    # debug info
    logging.debug(f"instantiate {builder.__name__}")
    for t in key_mapping:
        for k, v in key_mapping[t].items():
            string = f" {t:>10s}_args :  {k:>50s}"
            if k != v:
                string += f" <- {v:>50s}"
            logging.debug(string)
    logging.debug(f"...{builder.__name__}_param = dict(")
    logging.debug(f"...   optional_args = {final_optional_args},")
    logging.debug(f"...   positional_args = {positional_args})")

    try:
        instance = builder(**positional_args, **final_optional_args)
    except Exception as e:
        raise RuntimeError(
            f"Failed to build object with prefix `{prefix}` using builder `{builder.__name__}`"
        ) from e

    return instance, final_optional_args


def get_w_prefix(
        key: List[str],
        *kwargs,
        arg_dicts: List[dict] = [],
        prefix: Optional[Union[str, List[str]]] = [],
):
    """
    act as the get function and try to search for the value key from arg_dicts
    """

    # detect the input parameters needed from params
    config = Config(config={}, allow_list=[key])

    # sort out all possible prefixes
    if isinstance(prefix, str):
        prefix_list = [prefix]
    elif isinstance(prefix, list):
        prefix_list = prefix
    else:
        raise ValueError(f"prefix is with a wrong type {type(prefix)}")

    if not isinstance(arg_dicts, list):
        arg_dicts = [arg_dicts]

    # extract all the parameters that has the pattern prefix_variable
    # debug container to record all the variable name transformation
    key_mapping = {}
    for idx, arg_dict in enumerate(arg_dicts[::-1]):
        # fetch paratemeters that directly match the name
        _keys = config.update(arg_dict)
        key_mapping[idx] = {k: k for k in _keys}
        # fetch paratemeters that match prefix + "_" + name
        for idx, prefix_str in enumerate(prefix_list):
            _keys = config.update_w_prefix(
                arg_dict,
                prefix=prefix_str,
            )
            key_mapping[idx].update(_keys)

    # for logging only, remove the overlapped keys
    num_dicts = len(arg_dicts)
    if num_dicts > 1:
        for id_dict in range(num_dicts - 1):
            higher_priority_keys = []
            for id_higher in range(id_dict + 1, num_dicts):
                higher_priority_keys += list(key_mapping[id_higher].keys())
            key_mapping[id_dict] = {
                k: v
                for k, v in key_mapping[id_dict].items()
                if k not in higher_priority_keys
            }

    # debug info
    logging.debug(f"search for {key} with prefix {prefix}")
    for t in key_mapping:
        for k, v in key_mapping[t].items():
            string = f" {str(t):>10.10}_args :  {k:>50s}"
            if k != v:
                string += f" <- {v:>50s}"
            logging.debug(string)

    return config.get(key, *kwargs)



# -*- coding: utf-8 -*-

# @Time     : 2021/8/29 18:55
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx
"""Interface for ase.db and csv"""

import datetime
import json
import numbers
import os
import re
import warnings
from collections.abc import Callable
from typing import Dict, Union, Tuple, Any, List

import numpy as np
import pandas as pd
from ase.db import connect
from ase.db.core import Database, reserved_keys
from ase.db.row import AtomsRow, atoms2dict
from ase.io.jsonio import object_hook
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor

reserved_keys_extend = reserved_keys

special_kv_keys = {'project', 'name', 'symmetry', 'space_group'}

word = re.compile('[_a-zA-Z][_0-9a-zA-Z]*$')

_aaa = AseAtomsAdaptor()


def _code_dct(dct: Dict) -> Dict:
    """Change the numpy as list for store."""
    for k, v in dct.items():

        if hasattr(v, 'todict'):
            d = v.todict()

            if not isinstance(d, dict):
                raise RuntimeError('todict() of {} returned object of type {} '
                                   'but should have returned dict'
                                   .format(v, type(d)))
            if hasattr(v, 'ase_objtype'):
                d['__ase_objtype__'] = v.ase_objtype
            dct[k] = d
        elif isinstance(v, np.ndarray):
            flatobj = v.ravel()
            if np.iscomplexobj(v):
                flatobj.dtype = v.real.dtype
            dct[k] = {'__ndarray__': (v.shape,
                                      v.dtype.name,
                                      flatobj.tolist())}
        elif isinstance(v, np.integer):
            dct[k] = int(v)
        elif isinstance(v, np.bool_):
            dct[k] = bool(v)
        elif isinstance(v, datetime.datetime):
            dct[k] = {'__datetime__': v.isoformat()}
        elif isinstance(v, complex):
            dct[k] = {'__complex__': (v.real, v.imag)}

        # elif isinstance(v, np.ndarray):
        #     dct[k] = v.tolist()
        elif isinstance(v, dict):
            dct[k] = _code_dct(v)
        elif isinstance(v, (list, tuple)):
            dct[k] = [_code_dct(i) if isinstance(i, dict) else v for i in v]

        else:
            dct[k] = v

    return dct


def db_row2dct(row: AtomsRow
               ) -> Dict[str, Any]:
    """Convert row to dict,and change the numpy as list for store."""
    temp = {}
    temp.update(row.data)
    temp.update(row.key_value_pairs)
    dct = {i: row.__dict__[i] for i in reserved_keys if i in row.__dict__}
    temp.update(dct)

    temp = _code_dct(temp)

    return temp


def _decode_dct(dct: Dict) -> Dict:
    """resume the list to np.ndarray."""
    for k, v in dct.items():
        if k == "unique_id":
            dct[k] = v

        elif isinstance(v, float):
            if np.isnan(v):
                dct[k] = None
            else:
                dct[k] = v
        else:
            try:
                if isinstance(v, str):
                    v = eval(v)
                if isinstance(v, (list, tuple)):
                    if not v:
                        dct[k] = None
                    else:
                        try:
                            nda = np.array(v)
                            if nda.dtype in (np.float64, np.int32) and nda.ndim >= 1:
                                dct[k] = nda
                            else:
                                dct[k] = [_decode_dct(i) for i in v]
                        except:
                            dct[k] = [_decode_dct(i) for i in v]

                elif isinstance(v, dict):
                    v = object_hook(v)
                    if isinstance(v, dict):
                        dct[k] = _decode_dct(v)
                    else:
                        dct[k] = v
                else:
                    dct[k] = v
            except (ValueError, NameError, TypeError, SyntaxError):
                pass
    return dct


def collect_dct3(dct: Dict, kv_names=None) -> Tuple:
    """Take part the dict to 3 part."""
    if "data" in dct:
        data = dct["data"]
        del dct["data"]
    else:
        data = {}

    if "key_value_pairs" in dct:
        key_value_pairs = dct["key_value_pairs"]
        del dct["key_value_pairs"]
    else:
        key_value_pairs = {}

    temp = {}

    temp.update(data)
    temp.update(key_value_pairs)
    temp.update(dct)

    new_dct, new_data, new_kv = {}, {}, {}

    kv_keys = special_kv_keys if kv_names is None else special_kv_keys & set(kv_names)

    for k, v in temp.items():
        if k in reserved_keys:
            new_dct.update({k: v})
        elif k in kv_keys:
            new_kv.update({k: v})
        elif not word.match(k) or not isinstance(v, (numbers.Real, np.bool_)):
            new_data.update({k: v})
        else:
            new_kv.update({k: v})

    new_dct["key_value_pairs"] = new_kv
    new_dct["data"] = new_data

    return new_dct, new_kv, new_data


def db_dct2row(dct: Dict, kv_names=None) -> [AtomsRow, Dict, Dict]:
    """Convert dict to row."""
    dct = _decode_dct(dct)
    new_dct, new_kv, new_data = collect_dct3(dct, kv_names=kv_names)
    atomsrow = AtomsRow(new_dct)
    return atomsrow, new_kv, new_data


def collect_dct1(dct: Dict) -> Dict:
    new_dct, new_kv, new_data = collect_dct3(dct)
    new_data.update(new_kv)
    new_data.update(new_dct)
    return new_data


def db_to_csv(database: Union[str, Database], csv_file_name=""):
    """Store the db to csv."""
    if isinstance(database, str):
        database = connect(database)

    data = []

    for row in database.select(selection=None):
        try:
            data.append(db_row2dct(row))
        except (KeyError, StopIteration, AttributeError) as e:
            print(e, row)

    data = pd.DataFrame.from_dict(data)

    if csv_file_name:
        data.to_csv("{}.csv".format(os.path.splitext(csv_file_name)[0]))
    elif csv_file_name == "":
        csv_file_name = os.path.split(database.filename)[-1]
        csv_file_name = os.path.splitext(csv_file_name)[0]
        data.to_csv("{}.csv".format(csv_file_name))
    else:
        return data


def db_from_ase_csv(csv_file: Union[str, pd.DataFrame], new_database_name="new_database.db", index_col=0,
                    pop_name: List[str] = None
                    ) -> Database:
    """Convert csv(with strand Atoms table) to ase.db.
    see also: db_to_csv"""
    if pop_name is not None:
        assert isinstance(pop_name, list)
    if new_database_name == "" and isinstance(csv_file, str):
        new_database_name = "".join((os.path.splitext(csv_file)[0], ".db"))
    assert ".db" in new_database_name or ".json" in new_database_name
    if os.path.isfile(new_database_name):
        raise FileExistsError(new_database_name, "is exist, please change the `new_database_name`.")
    database = connect(new_database_name)
    if isinstance(csv_file, str):
        file = pd.read_csv(csv_file, index_col=index_col)
    else:
        file = csv_file

    data = file.to_dict(orient="index")

    with database:
        for k in data.keys():
            atomsrow, new_kv, new_data = db_dct2row(data[k])
            if pop_name is not None:
                for i in pop_name:
                    if i in new_kv:
                        del new_kv[i]
                    elif i in new_data:
                        del new_data[i]
            database.write(atomsrow, key_value_pairs=new_kv, data=new_data)
    return database


def db_from_structure_csv(csv_file: Union[str, pd.DataFrame], new_database_name="new_database.db", index_col=0,
                          structure_index_name: str = None, fmt: Union[str, Callable] = "json",
                          pop_name: List[str] = None
                          ):
    """Convert csv to ase.db. The structure must be offered."""

    if isinstance(csv_file, str):
        file = pd.read_csv(csv_file, index_col=index_col)
    else:
        file = csv_file

    data = file.to_dict(orient="index")

    if new_database_name == "" and isinstance(csv_file, str):
        new_database_name = "".join((os.path.splitext(csv_file)[0], ".db"))
    assert ".db" in new_database_name or ".json" in new_database_name

    if structure_index_name and structure_index_name in file.columns:
        name = structure_index_name
    elif "structure" in file.columns:
        name = "structure"
    else:
        raise NameError("There must be structure index name.")

    database = db_from_structure_dict(data, new_database_name=new_database_name,
                                      structure_index_name=name, fmt=fmt, pop_name=pop_name)

    return database


def db_from_structure_json(json_file: Union[str, dict], new_database_name="new_database.db",
                           structure_index_name: str = "structure", fmt: Union[str, Callable] = "json",
                           pop_name: List[str] = None
                           ):
    """Convert csv to ase.db. The structure must be offered."""

    def read_json(i):
        f = open(i)
        entries = json.load(f)
        f.close()
        return entries

    if isinstance(json_file, str):
        data = read_json(json_file)
    else:
        data = json_file

    if new_database_name == "" and isinstance(json_file, str):
        new_database_name = "".join((os.path.splitext(json_file)[0], ".db"))
    assert ".db" in new_database_name or ".json" in new_database_name

    if structure_index_name is None:
        structure_index_name = "structure"
    name = structure_index_name

    database = db_from_structure_dict(data, new_database_name=new_database_name,
                                      structure_index_name=name, fmt=fmt, pop_name=pop_name)

    return database


# def fmt(st):
#     st.lattice._pbc=np.array([True,True,True])
#     try:
#         atoms = aaa.get_atoms(st)
#     except ValueError:
#         st.remove_site_property("selective_dynamics")
#         atoms = aaa.get_atoms(st)
#     return atoms


def db_from_structure_dict(data, new_database_name="new_database.db",
                           structure_index_name: str = None, fmt: Union[str, Callable] = "json",
                           pop_name: List[str] = None):
    """Convert csv to ase.db. The structure must be offered."""
    if pop_name is not None:
        assert isinstance(pop_name, list)

    assert ".db" in new_database_name or ".json" in new_database_name
    if os.path.isfile(new_database_name):
        raise FileExistsError(new_database_name, "is exist, please change the `new_database_name`.")

    if isinstance(data, list):
        number = range(len(data))
    else:
        number = data.keys()

    database = connect(new_database_name)

    with database:
        for k in number:
            try:
                datak = data[k]
                structure = datak.pop(structure_index_name)
                if isinstance(fmt, str):
                    st = Structure.from_str(structure, fmt=fmt)
                    atoms = _aaa.get_atoms(st)
                elif isinstance(fmt, Callable):
                    atoms = fmt(structure)
                dct = _decode_dct(datak)
                new_dct, new_kv, new_data = collect_dct3(dct)
                dct = atoms2dict(atoms)
                dct.update(new_dct)
                if pop_name is not None:
                    for i in pop_name:
                        if i in new_kv:
                            del new_kv[i]
                        elif i in new_data:
                            del new_data[i]

                database.write(dct, key_value_pairs=new_kv, data=new_data)
            except NameError as e:
                # except BaseException as e:
                warnings.warn("The {} sample is can't be analysis.".format(k))
                print(e)

    return database



# -*- coding: utf-8 -*-

# @Time    : 2021/8/1 18:07
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
import copy
from typing import Union, List

import numpy as np
import torch
from torch.nn import Module
from torch_geometric.data import InMemoryDataset, DataLoader, Data

from hpc2ml.data.dataset import SimpleDataset


def print_log_dataset(dataset: Union[InMemoryDataset, SimpleDataset, List[Data], Data]):
    nfeat_node = -1
    nfeat_edge = -1
    nfeat_state = -1

    if isinstance(dataset, List):
        dataset = dataset[0]
    if isinstance(dataset, Data):
        data = dataset
    else:
        data = dataset.data

    assert isinstance(data, Data)

    print("\n数据信息:\n###############")

    try:
        print("图（化合物）数={},".format(max(data.idx) + 1))
    except AttributeError:
        pass
    try:
        nfeat_node = data.x.shape
        print("原子（节点）总个数: {}, 原子特征数: {}".format(nfeat_node[0], nfeat_node[1]))
    except AttributeError:
        pass

    try:
        nfeat_edge = data.edge_attr.shape

        print("键（连接）连接个数: {}, 键特征数: {}".format(nfeat_edge[0], nfeat_edge[1]))
    except AttributeError:
        pass

    try:
        nfeat_state = data.state_attr.shape

        print("状态数: {}, 状态特征数 {}".format(nfeat_state[0], nfeat_state[1]))
    except AttributeError:
        pass

    print("\n建议参数如下(若后处理，以处理后为准):")
    if nfeat_node != -1:
        print("nfeat_node={},".format(nfeat_node[1]))
    if nfeat_edge != -1:
        print("nfeat_edge={},".format(nfeat_edge[1]))
    if nfeat_state != -1:
        print("nfeat_state={},".format(nfeat_state[1]))


def print_log_dataloader(dataloader: DataLoader, print_index_of_sample=False):
    dataloader0 = copy.copy(dataloader)
    names = ['x', 'edge_attr', 'energy', 'pos', 'batch', 'ptr', 'z', 'idx', 'state_attr', 'adj_t', 'edge_weight',
             "num_graphs"]
    for data in dataloader0:
        shapes = []
        for i in names:
            shapes.append(getattr(data, i, None))

        print("\n每批信息（示例第0批）:\n###############")

        if shapes[-1] is not None:
            print("图（化合物）数={},".format(np.array(shapes[-1])))
        if shapes[0] is not None:
            print("节点（原子）数={}, 原子特征数={}".format(shapes[0].shape[0], shapes[0].shape[1]))
        if shapes[1] is not None:
            print("键（连接）数={}, 键特征数={}".format(shapes[1].shape[0], shapes[1].shape[1]))
        if shapes[8] is not None:
            print("状态数={}, 状态特征数={}".format(shapes[8].shape[0], shapes[8].shape[1]))

        if print_index_of_sample:
            if shapes[7] is not None:
                print("样本序号={},".format(np.array(shapes[7])))

        print("\n建议参数如下:")
        if shapes[0] is not None:
            print("nfeat_node={},".format(shapes[0].shape[1]))
        if shapes[1] is not None:
            print("nfeat_edge={},".format(shapes[1].shape[1]))
            print("num_edge_gaussians={},".format(shapes[1].shape[1]))
        if shapes[8] is not None:
            print("nfeat_state={},".format(shapes[8].shape[1]))

        break


def make_dot_(var, params=None, show_attrs=False, show_saved=False, max_attr_chars=50):
    # 画出 PyTorch 自动梯度图 autograd graph 的 Graphviz 表示.
    # 蓝色节点表示有梯度计算的变量Variables;
    # 橙色节点表示用于 torch.autograd.Function 中的 backward 的张量 Tensors.
    """
    Produces Graphviz representation of PyTorch autograd graph.

    First install graphviz:
        https://forum.graphviz.org/t/new-simplified-installation-procedure-on-windows/224  (windows)

        yum install graphviz (centos)

        apt-get install graphviz (ubuntu)

    Second:
        pip install graphviz
        pip install torchviz

    use:
        >>> vis_graph = make_dot_(y_pred, params=dict(list(model.named_parameters())))
        >>> vis_graph.render(format="pdf")

    If a node represents a backward function, it is gray. Otherwise, the node
    represents a tensor and is either blue, orange, or green:

     - Blue: reachable leaf tensors that requires grad (tensors whose `.grad`
         fields will be populated during `.backward()`)
     - Orange: saved tensors of custom autograd functions as well as those
         saved by built-in backward nodes
     - Green: tensor passed in as outputs
     - Dark green: if any output is a view, we represent its base tensor with
         a dark green node.

    Args:
        var: output tensor
        params: dict of (name, tensor) to add names to node that requires grad
            params=dict(model.named_parameters()
        show_attrs: whether to display non-tensor attributes of backward nodes
            (Requires PyTorch version >= 1.9)
        show_saved: whether to display saved tensor nodes that are not by custom
            autograd functions. Saved tensor nodes for custom functions, if
            present, are always displayed. (Requires PyTorch version >= 1.9)
        max_attr_chars: if show_attrs is `True`, sets max number of characters
            to display for any given attribute.
    """
    from torchviz import make_dot
    return make_dot(var, params=params, show_attrs=show_attrs, show_saved=show_saved, max_attr_chars=max_attr_chars)


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name):
        self.name = name
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def record(self, val, n=1):
        self.val = val
        self.sum = self.sum + val * n  # (if pytorch not support +=)
        self.count += n
        self.avg = self.sum / self.count

    def text(self):
        return ' {loss.name} {loss.val:.4f} ({loss.avg:.4f}) '.format(loss=self)


class AverageMeterTotal(dict):

    def __init__(self, *names):
        super(AverageMeterTotal, self).__init__()
        for i in names:
            self.update({i: AverageMeter(i)})

    def reset(self):
        [v.reset() for k, v in self.items()]

    def text(self):
        return "".join([vi.text() for vi in self.values()])

    def record(self, name, val, n=1):
        self[name].record(val, n)


def get_layers_with_weight(model: Module, add_name=""):
    """Get layers with weight. use get_target_layer function ranther than this is suggested."""
    name_weight_module = {}
    for name, layer in model._modules.items():
        name = "{}->{}".format(add_name, name)
        # if isinstance(layer,Module):
        if len(layer._modules) > 0:
            name_weight_module.update(get_layers_with_weight(layer, name))
        else:
            if hasattr(layer, "weight"):
                name_weight_module[name] = layer
            else:
                pass
    return name_weight_module


def get_target_layer(model, target_layer) -> dict:
    """which layer to hook.  run model.named_modules() to get the name of layer.
    """
    mod = {}
    if isinstance(target_layer, (list, tuple)) and len(target_layer) > 0:

        for i in target_layer:
            for name, module in model.named_modules():
                if name == i or module is i:
                    mod.update({name: module})
                    break
    elif isinstance(target_layer, str):

        if target_layer == "all":
            # all layer
            for name, module in model.named_modules():
                mod.update({name: module})
        elif target_layer == "top":
            # top layer
            mod.update(model._modules)
        elif target_layer == "weight":
            # weight layer
            for name, module in model.named_modules():
                if hasattr(module, "weight"):
                    mod[name] = module
        elif target_layer == "my_weight":
            # weight layer
            mmdict = get_layers_with_weight(model)
            mod.update(mmdict)

    else:

        mmdict = get_layers_with_weight(model)
        mod.update(mmdict)

    return mod


class LogModule:
    """Get message of module."""

    def __init__(self, model, target_layer="weight"):
        self._stats_log = []
        self.stats_log_i = {}
        self.stats_log_format = {}

        self.layer_dict = get_target_layer(model, target_layer)

    def _get_weight(self):
        name_weight_module = {}
        for name, layer in self.layer_dict.items():
            if hasattr(layer, "weight"):
                v = layer.weight.detach().cpu().numpy().copy()
                name_weight_module[name] = np.array([np.mean(v, axis=0), np.std(v, axis=0)])
        return name_weight_module

    def record(self, append=True):
        stats_log_i = self._get_weight()
        if append:
            self._stats_log.append(stats_log_i)
            self.stats_log_i = self._stats_log[-1]
        else:
            self.stats_log_i = stats_log_i

    def stats_single(self):
        return self.stats_log_i

    def stats_loop(self):

        result = {}

        for i in self._stats_log[0].keys():
            nps = []
            for n, v in enumerate(self._stats_log):
                nps.append(v[i])

            result[i] = np.array(nps)

        self.stats_log_format = result

        return result


class HookGradientLayer:
    """This is used for one sample to check which is the import message for this target."""

    def __init__(self, layer_name, layer):
        # register a hook to save values of activations and gradients
        self.activations = None
        self.gradients = None
        self.layer_name = layer_name
        self._stats_log = []
        self.stats_log_i = {}
        self.stats_log_format = {}

        self.forward_hook = layer.register_forward_hook(self.hook_fn_act)
        self.backward_hook = layer.register_full_backward_hook(self.hook_fn_grad)
        # self.backward_hook = layer.register_backward_hook(self.hook_fn_grad)

    def hook_fn_act(self, module, input, output):
        self.activations = output

    def hook_fn_grad(self, module, grad_input, grad_output):
        self.gradients = grad_output[0]

    def remove(self):
        self.forward_hook.remove()
        self.backward_hook.remove()

    def _get_weight(self):
        try:
            activations = self.activations
            gradients = self.gradients
            v = (gradients * activations)
            v = v.detach().cpu().numpy()
            if v.ndim >= 2:
                return {self.layer_name: np.array([np.mean(v, axis=0), np.std(v, axis=0)])}
            else:
                return {self.layer_name: np.array([v, np.zeros_like(v)])}
        except:
            return {}

    def record(self, append=True):
        stats_log_i = self._get_weight()
        if append:
            self._stats_log.append(stats_log_i)
            self.stats_log_i = self._stats_log[-1]
        else:
            self.stats_log_i = stats_log_i

    def stats_single(self):
        return self.stats_log_i

    def stats_loop(self):
        result = {}
        for i in self._stats_log[0].keys():
            nps = []
            for n, v in enumerate(self._stats_log):
                nps.append(v[i])

            result[i] = np.array(nps)

        self.stats_log_format = result

        return result


class HookGradientModule():
    def __init__(self, model, target_layer="all"):
        self.stats_log_format = []
        self.stats_log_i = {}

        self.layer_dict = get_target_layer(model, target_layer)

        self.svls = []
        for k, v in self.layer_dict.items():
            svl = HookGradientLayer(k, v)
            self.svls.append(svl)

    def apply(self, func, **kwargs):
        result = []
        for i in self.svls:
            ri = getattr(i, func)(**kwargs)
            result.append(ri)
        return result

    def record(self, append=True):
        self.apply("record", append=append)

    def stats_single(self):
        self.stats_log_i = {}
        results = self.apply("stats_single")
        [self.stats_log_i.update(i) for i in results]
        return self.stats_log_i

    def stats_loop(self):
        self.stats_log_format = {}
        results = self.apply("stats_loop")
        [self.stats_log_format.update(i) for i in results]
        return self.stats_log_format


# def for_hook(module, input, output):
#     print(module)
#     for val in input:
#         print("input val:", val)
#     for out_val in output:
#         print("output val:", out_val)

def print_cuda_usage():
    print("Memory Allocated:", torch.cuda.memory_allocated() / (1024 * 1024))
    print(
        "Max Memory Allocated:",
        torch.cuda.max_memory_allocated() / (1024 * 1024),
    )
    print("Memory Cached:", torch.cuda.memory_cached() / (1024 * 1024))
    print("Max Memory Cached:", torch.cuda.max_memory_cached() / (1024 * 1024))



import os
import shutil
import sys
import time
import warnings
from typing import Union, Tuple, List, Optional

import numpy as np
import torch
from sklearn import metrics
from torch.nn import Module
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import DataLoader

from hpc2ml.utils.log_model import LogModule, HookGradientModule, make_dot_, AverageMeterTotal


def class_eval(prediction, target):
    """Classification."""
    prediction = np.exp(prediction.numpy())
    target = target.numpy()
    pred_label = np.argmax(prediction, axis=1)
    target_label = np.squeeze(target)
    if not target_label.shape:
        target_label = np.asarray([target_label])
    if prediction.shape[1] == 2:
        precision, recall, fscore, _ = metrics.precision_recall_fscore_support(
            target_label, pred_label, average='binary')
        auc_score = metrics.roc_auc_score(target_label, prediction[:, 1])
        accuracy = metrics.accuracy_score(target_label, pred_label)
    else:
        raise NotImplementedError
    return accuracy, precision, recall, fscore, auc_score


def mae(prediction, target):
    """
    Computes the mean absolute error between prediction and target

    Parameters
    ----------
    prediction: torch.Tensor (N, 1)
    target: torch.Tensor (N, 1)
    """
    return torch.mean(torch.abs(target - prediction))


class ProcessOutLabel():
    def __init__(self, multi_loss=False, target_name=("energy", "forces", 'stress'),
                 process_label=None, process_out=None,
                 ):

        self.multi_loss = multi_loss

        self.target_name = target_name

        if target_name is None:
            target_name = ("energy", "forces", "stress")

        self.target_name = tuple(target_name) if isinstance(target_name, (tuple, list)) else (
            target_name,)

        if not self.multi_loss:
            self.target_name = [self.target_name[0], ]
            self.target_number = 1
        else:
            self.target_number = len(target_name)

        def func11(data):
            return tuple([data[i] for i in self.target_name])

        if process_label is None:
            self._process_label = func11
        else:
            self._process_label = process_label

        def func2(y_or_ys, data=None):
            if isinstance(y_or_ys, dict):
                return tuple([y_or_ys[i] for i in self.target_name])
            else:
                return y_or_ys

        if process_out is None:
            self._process_out = func2
        else:
            self._process_out = process_out

    def process_out(self, y_or_ys, data=None):
        y_or_ys = self._process_out(y_or_ys, data)

        if isinstance(y_or_ys, torch.Tensor):
            return y_or_ys

        if len(y_or_ys) < self.target_number:
            raise KeyError(f"The target number {self.target_number} is not"
                           f" consist with output {len(y_or_ys)}. "
                           f"check `multi_loss` and `target_name`")

        if len(y_or_ys) > self.target_number:
            warnings.warn(f"The target number {self.target_number} less than"
                          f" the output {len(y_or_ys)}. "
                          f"Please defined your loss_method, and neglect the redundant output-i."
                          f"Or add your `target_name`", UserWarning)
        return y_or_ys if len(y_or_ys) > 1 else y_or_ys[0]

    def process_label(self, data):
        label = self._process_label(data=data)

        if isinstance(label, torch.Tensor):
            return label

        return label if len(label) > 1 else label[0]


class LearningFlow:
    """
    LearningFlow for training.
    
    Examples:

        >>> test_dataset = dataset[:1000]
        >>> val_dataset = dataset[1000:2000]
        >>> train_dataset = dataset[2000:3000]
        >>> import torch_geometric.transforms as T
        >>> train_dataset = SimpleDataset(data_train, pre_transform=T.ToSparseTensor())
        >>> test_dataset = SimpleDataset(data_test,pre_transform=T.ToSparseTensor())
        >>> val_dataset = SimpleDataset(val_data,pre_transform=T.ToSparseTensor())

        >>> train_loader = DataLoader(
        ... dataset=train_dataset,
        ... batch_size=200,
        ... shuffle=False,
        ... num_workers=0)

        >>> test_loader = ...
        >>> val_loader = ...


        >>>  model = CrystalGraphConvNet(nfeat_node=91,
        ...  nfeat_edge=3,
        ...  nfeat_state=29,
        ...  nc_node_hidden=128,
        ...  nc_node_interaction=64,
        ...  num_interactions=2,)
        >>> # model = CrystalGraphGCN(...)
        >>> # model = CrystalGraphGCN2(...)
        >>> # model = CrystalGraphGAT(...)
        >>> # model = SchNet(...)
        >>> # model = MEGNet(...)
        >>> # model = SchNet(...)


        >>> optimizer = torch.optim.Adam(model.parameters(), lr=0.01)
        >>> scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.7, patience=2,...min_lr=0.001)
        >>> device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')

        >>> lf= LearningFlow(model, train_loader, test_loader=val_loader, device= "cuda:1",
        ... optimizer=None, clf= False, loss_method=None, learning_rate = 1e-3,
        ... weight_decay= 0.01, checkpoint=True, scheduler=scheduler,
        ... loss_threshold= 0.1, print_freq= None, print_what="all")

        >>> lf.run(50)

    where the dataset could from StructureToData or :class:`featurebox.test_featurizers.base_graph_geo.StructureGraphGEO`
    """

    def __init__(self, model: Module, train_loader: DataLoader, test_loader: DataLoader,
                 device: Union[str, object] = "cpu",
                 optimizer=None, clf: bool = False, loss_method=None, learning_rate: float = 1e-3,
                 weight_decay: float = 0.0, checkpoint=True, scheduler=None, debug: str = None,
                 target_layers: Union[str, List, Tuple] = "all", multi_loss=False,
                 target_name: Optional[Tuple] = ("energy", "forces",),
                 loss_threshold: float = 0.1, print_freq: Union[int, str, None] = 10, print_what="all",
                 process_label=None, process_out=None, note=None, store_filename='checkpoint.pth.tar'):
        """

        Parameters
        ----------
        model: module
        train_loader: DataLoader
        test_loader: DataLoader
        device:str
            such as "cuda:0","cpu"
        optimizer:torch.Optimizer
        clf:bool
            under exploit......
        loss_method:torch._Loss
            1. In default, for classification, loss_method = torch.nn.CrossEntropyLoss().
            And for claffication, torch.nn.MSELoss(), see more in torch
            2. If multi_loss, use function to deal with multi-output and return one single score. such has:
            >>> def mlm(ys_pred, ys_true):
            ...     mse = torch.potlayer.MSELoss()
            ...     ls = []
            ...     for i,j in zip(ys_pred, ys_true):
            ...           ls.append(mse(i,j))
            ...     return torch.mean(ls)
        multi_loss:bool
            for multi-optput metwork.
        target_name:tuple of str
            other names as target in data, except 'y'.
        learning_rate:float
            see more in torch
        weight_decay:float
            see more in torch
        checkpoint:bool
            save checkpoint or not.
        loss_threshold:
            see more in torch
        print_freq:int
            print frequency
        print_what:str
            "all","train","test" log.
        scheduler:
            scheduler, see more in torch
        process_label:Callable,
            function to get true y/label, mainly change shape. input: Data, output: List[Tensor].
            default:
            >>> def func11(data):
            ...    return tuple([data[i] for i in self.target_name])

        process_out:Callable
            function to get predict y, mainly change shape.
            input:(Union[Tensor,List[Tensor], Data), output: List[Tensor].
            default:
            >>> def func2(y_or_ys, data=None):
            >>> if isinstance(y_or_ys, dict):
            ...    return tuple([y_or_ys[i] for i in self.target_name])
            >>> else:
            ...    return y_or_ys
        store_filename:str
            filename to store.
        """

        if note is None:
            note = {}
        self.note = note

        self.train_loader = train_loader
        self.test_loader = test_loader

        device = torch.device(device)

        self.device = device
        self.model = model
        self.model.to(device)
        self.clf = clf
        self.loss_method = loss_method
        self.optimizer = optimizer
        self.checkpoint = checkpoint
        self.store_filename = store_filename

        self.weight_log = LogModule(model=model, target_layer="weight")

        self.debug = "" if debug is None else debug
        if "hook" in self.debug or "loop" in self.debug:
            self.gradient_log = HookGradientModule(model=model, target_layer=target_layers)

        self.train_batch_number = len(self.train_loader)

        if print_freq is None:
            self.print_freq = self.train_batch_number
        else:
            self.print_freq = self.train_batch_number if not isinstance(print_freq, int) else print_freq

        if self.optimizer is None or self.optimizer == "Adam":
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr=learning_rate, weight_decay=weight_decay)
        elif self.optimizer == "SGD":
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=learning_rate, weight_decay=weight_decay)
            # L2 regularization
        else:
            self.optimizer = optimizer

        self.multi_loss = multi_loss

        if self.loss_method is None:
            if self.clf is True:
                lm = torch.nn.CrossEntropyLoss()
            elif self.clf == "multi_label":
                lm = torch.nn.L1Loss()
                # 主要是用来判定实际的输出与期望的输出的接近程度 MAE: 1/N |y_pred-y| y 为多列
            else:
                lm = torch.nn.MSELoss()

            if not self.multi_loss:
                self.loss_method = lm
            else:
                def mlm(y_preds: tuple, batch_ys: tuple):
                    ls = []
                    for i, j in zip(y_preds, batch_ys):
                        ls.append(lm(i, j))
                    return sum(ls)

                warnings.warn("The default 'loss_method' for 'multi_loss' is just add all loss. which is simple, "
                              "we suggest use your own 'loss_method' !!!", UserWarning)
                self.loss_method = mlm

        else:
            self.loss_method = loss_method

        if scheduler is None:
            self.scheduler = ReduceLROnPlateau(self.optimizer, mode='min', factor=0.5, patience=7, min_lr=0.00001)
        else:
            self.optimizer = scheduler.optimizer
            self.scheduler = scheduler
        self.best_error = 1000000.0
        self.threshold = loss_threshold
        # *.pth.tar or str
        self.run_train = self.run
        self.fit = self.run_train
        self.print_what = print_what

        if self.clf and process_label is None:
            warnings.warn("We suggest defined the process_label for classification problem.")

        self.pol = ProcessOutLabel(multi_loss=self.multi_loss, target_name=target_name,
                                   process_out=process_out, process_label=process_label)

        self.start_epoch = 0

        if self.clf:
            self.meters = AverageMeterTotal("Time", "losses", "accuracies", "precisions", "recalls", "fscores",
                                            "auc_scores")
        else:
            self.meters = AverageMeterTotal("Time", "losses", "mae_errors")

        self.single_print = True

    def my_debug(self):
        # debug

        if "loop" in self.debug:
            if "weight" in self.debug:
                self.weight_log.stats_loop()
            if "hook" in self.debug:
                self.gradient_log.stats_loop()

        if "weight" in self.debug:
            # just for 'single',
            self.model.train()
            for m, data in enumerate(self.train_loader):  # just record for once,
                data = data.to(self.device)
                self.model(data)
                self.weight_log.record(append=False)
                break
            self.weight_log.stats_single()

        if "graphviz" in self.debug:
            # just for 'single',
            self.model.train()
            for m, data in enumerate(self.train_loader):
                data = data.to(self.device)
                y_pred = self.model(data)
                vis_graph = make_dot_(y_pred, params=dict(list(self.model.named_parameters())))
                vis_graph.render(format="pdf", view=False)
                break

        if "hook" in self.debug:
            # hook just for 'single'
            # which layer to hook.
            self.model.train()
            for m, data in enumerate(self.train_loader):  # just record for once,
                data = data.to(self.device)
                batch_y = self.process_label(data=data)
                self.model.zero_grad()
                y_pred = self.model(data)
                y_pred = self.process_out(y_pred)
                lossi = self.loss_method(y_pred, batch_y)
                lossi.backward(retain_graph=True)
                # lossi.backward()
                self.gradient_log.record(append=False)
                break
            self.gradient_log.stats_single()

    def process_label(self, data=None):
        return self.pol.process_label(data=data)

    def process_out(self, y, data=None):
        return self.pol.process_out(y, data)

    def save_checkpoint(self, state, is_best, store_filename):
        if store_filename is not None:
            self.store_filename = store_filename
        torch.save(state, self.store_filename)
        if is_best:
            if self.store_filename != "model_best.pth.tar":
                shutil.copyfile(self.store_filename, 'model_best.pth.tar')

    def run(self, epoch=50, warm_start: bool = False, store_filename=None):
        """
        run loop.

        Parameters
        ----------
        epoch:int
            epoch.
        warm_start: bool
            warm start or not.
        store_filename: str
            The name of resume file, 'checkpoint.pth.tar' or 'model_best.pth.tar'
            If warm_start, try to resume from local disk.

        """
        if store_filename is not None:
            self.store_filename = store_filename
        else:
            self.store_filename = 'checkpoint.pth.tar'

        if warm_start:
            resume = self.store_filename
            if os.path.isfile(resume):
                print("=> loading checkpoint '{}'".format(resume))
                checkpoint = torch.load(resume)
                self.start_epoch = checkpoint['epoch']
                self.best_error = checkpoint['best_error']
                self.model.load_state_dict(checkpoint['state_dict'])
                self.optimizer.load_state_dict(checkpoint['optimizer'])
                self.note = checkpoint['note']
                print("=> loaded checkpoint '{}' (epoch {})"
                      .format(resume, checkpoint['epoch']))
            else:
                print("=> no checkpoint found at '{}'".format(resume))

        if self.start_epoch > 0:
            print("Try to run start from 'resumed epoch' {} to 'epoch' {}".format(self.start_epoch,
                                                                                  epoch + self.start_epoch))
            if epoch == 0:
                print("No new epoch added.")

        for epochi in range(self.start_epoch, epoch + self.start_epoch):

            self._train(epochi)

            if "loop" in self.debug:  # record each epochi
                self.weight_log.record()
                self.gradient_log.record()

            score = self._validate(epochi)

            if score != score:
                print('Exit due to NaN output value.')
                sys.exit(1)
            try:
                self.scheduler.step(score)
            except TypeError:
                self.scheduler.step()

            is_best = score < self.best_error
            self.best_error = min(score, self.best_error)

            if self.checkpoint:
                self.save_checkpoint({
                    'epoch': epochi + 1,
                    'state_dict': self.model.state_dict(),
                    'best_error': score,
                    'optimizer': self.optimizer.state_dict(),
                    'note': self.note
                }, is_best, store_filename=self.store_filename)

            if score <= self.threshold:
                print("Up to requirements and early termination in epoch ({})".format(epochi))
                break

        self.my_debug()
        self.start_epoch = self.start_epoch + epoch

    def _train(self, epochi):
        self.model.train()
        self.meters.reset()

        for m, data in enumerate(self.train_loader):

            point = time.time()
            data = data.to(self.device)
            batch_y = self.process_label(data=data)

            self.optimizer.zero_grad()

            y_pred = self.model(data)

            y_pred = self.process_out(y_pred, data)

            lossi = self.loss_method(y_pred, batch_y)

            lossi.backward()

            self.optimizer.step()

            del data

            # ###########

            self.print_rec(lossi, y_pred, batch_y, point)

            if self.print_freq != self.train_batch_number and m % self.print_freq == 0 \
                    and self.print_what in ["all", "train"]:
                print('Train: [{0}][{1}/{2}] {3}'.format(epochi, m, len(self.train_loader), self.meters.text()))

        if self.print_freq == self.train_batch_number:
            print('Train: [{}] {}'.format(epochi, self.meters.text()))

    def print_rec(self, lossi, y_pred, data_y, point):

        if isinstance(y_pred, tuple):
            y_pred = y_pred[0]
        if isinstance(data_y, tuple):
            data_y = data_y[0]

        batch_size = data_y.size(0)

        lossi = float(lossi.__repr__()[7:].split(",")[0])  # more speed?
        self.meters['losses'].record(lossi, batch_size)

        if self.clf is False:
            mae_error = mae(y_pred, data_y)
            mae_error = float(mae_error.__repr__()[7:].split(",")[0])  # more speed?

            self.meters['mae_errors'].record(mae_error, batch_size)

        else:
            accuracy, precision, recall, fscore, auc_score = \
                class_eval(y_pred.detach().cpu(), data_y.cpu())
            self.meters['accuracies'].record(accuracy, batch_size)
            self.meters['precisions'].record(precision, batch_size)
            self.meters['recalls'].record(recall, batch_size)
            self.meters['fscores'].record(fscore, batch_size)
            self.meters['auc_scores'].record(auc_score, batch_size)

        time_range = time.time() - point
        self.meters['Time'].record(time_range)

    def _validate(self, epochi):
        self.model.eval()
        self.meters.reset()

        for m, data in enumerate(self.test_loader):
            point = time.time()

            data = data.to(self.device)
            batch_y = self.process_label(data=data)
            y_pred = self.model(data)
            y_pred = self.process_out(y_pred, data)

            lossi = self.loss_method(y_pred, batch_y)

            del data

            # ###########

            self.print_rec(lossi, y_pred, batch_y, point)

            if self.print_freq != self.train_batch_number and m % self.print_freq == 0 \
                    and self.print_what in ["all", "train"]:
                print('Test: [{0}][{1}/{2}] {3}'.format(epochi, m, len(self.train_loader), self.meters.text()))

        if self.print_freq == self.train_batch_number:
            print('Test: [{}] {}'.format(epochi, self.meters.text()))

        if self.clf is False:
            return self.meters['mae_errors'].avg
        else:
            return self.meters['auc_scores'].avg

    def score(self, predict_loader, scoring=None):
        """Return ."""
        y_pre, y_true = self.predict(predict_loader)
        if not scoring:
            scoring = self.loss_method
        return float(scoring(y_pre, y_true))

    def predict(self, predict_loader: DataLoader, return_y_true=False,
                add_hook=False, hook_layer=None, device='cpu'):
        """
        Just predict by model,and add one forward hook to get processing output.

        Parameters
        ----------
        predict_loader:DataLoader
            MGEDataLoader, the target_y could be ``None``.
        return_y_true:bool
            if return_y_true, return (y_preds, y_true)
        add_hook:bool
        hook_layer:
            one layer of model. could get layers by ``get_target_layer`` (the values).

        Returns
        -------
        y_pred:tensor
        y_true:tensor
            if return_y_true

        """

        handles = []
        if add_hook:
            try:
                self.forward_hook_list = []

                def for_hook(module, input, output):
                    self.forward_hook_list.append(output.detach().cpu())

                handles.append(hook_layer.register_forward_hook(for_hook))
            except BaseException as e:
                print(e)
                raise AttributeError("use ``hook_layer`` to defined the hook layer.")

        res = simple_predict(self.model, predict_loader, return_y_true=return_y_true, device=device,
                             process_out=self.pol.process_out, process_label=self.pol._process_label,
                             multi_loss=self.multi_loss)



        if add_hook:
            [i.remove() for i in handles]  # del

        return res


def cat(y_preds):
    if isinstance(y_preds[0], (tuple, list)):
        y_preds = list(zip(*y_preds))
        if len(y_preds) == 1:
            y_preds = [i[0] for i in y_preds]
        else:
            y_preds = [torch.cat(i, dim=0) for i in y_preds]
    else:
        if len(y_preds) == 1:
            y_preds = y_preds[0]
        else:
            y_preds = torch.cat(y_preds, dim=0)
    return y_preds


def simple_predict(model, predict_loader: DataLoader, return_y_true=False, device='cpu', process_out=None,
                   process_label=None, target_name=("energy", "forces", "stress"), multi_loss=False):
    """
    Just predict by model,and add one forward hook to get processing output.

    Parameters
    ----------
    model:torch.nn.Module
        network.
    predict_loader:DataLoader
        MGEDataLoader, the target_y could be ``None``.
    return_y_true:bool
        if return_y_true, return (y_preds, y_true)

    Returns
    -------
    y_pred:tensor
    y_true:tensor
        if return_y_true

    """

    pol = ProcessOutLabel(multi_loss=multi_loss, target_name=target_name,
                          process_label=process_label, process_out=process_out)

    model.eval()
    model.to(device)

    ############

    y_preds = []
    y_true = []
    for data in predict_loader:
        data = data.to(device)

        y_prei = model(data)
        if isinstance(y_prei, tuple):
            y_prei = [i.detach() for i in y_prei]
        else:
            y_prei = y_prei.detach()

        y_preds.append(pol.process_out(y_prei, data))
        if return_y_true:
            y_true.append(pol.process_label(data))

    y_preds = cat(y_preds)

    if return_y_true:
        y_true = cat(y_true)
        return y_preds, y_true
    else:
        return y_preds


def load_check_point(model, resume_file="checkpoint.pth.tar", optimizer=None, device="cpu"):
    """
    load check_point without flow, just for predict

    Parameters
    ----------
    model:
        network
    resume_file:
        'model_best.pth.tar' or 'checkpoint.pth.tar'
    optimizer:
        optimizer

    Returns
    -------
    model, optimizer, start_epoch, best_error

    """
    print("=> loading checkpoint '{}'".format(resume_file))
    checkpoint = torch.load(resume_file, map_location=torch.device(device))
    start_epoch = checkpoint['epoch']
    best_error = checkpoint['best_error']
    note = checkpoint['note']

    if model is not None:
        model.load_state_dict(checkpoint['state_dict'])
    else:
        model = None

    if optimizer is not None:
        optimizer.load_state_dict(checkpoint['optimizer'])
    else:
        optimizer = None

    print("=> loaded checkpoint '{}' (epoch {}) with score {}"
          .format(resume_file, start_epoch, best_error))
    return model, optimizer, start_epoch, best_error, note



import random
from collections import OrderedDict
from typing import Dict, Tuple, Callable, Any, Sequence, Union, Optional

import torch
from e3nn import o3

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST
from hpc2ml.utils.nequliputils.auto_init import instantiate


class GraphModuleMixin:
    r"""Mixin parent class for ``torch.nn.Module``s that act on and return ``AtomicDataDict.ST`` \ ``AtomicDataDict.Type`` graph data.

    All such classes should call ``_init_irreps`` in their ``__init__`` functions with information on the
    data fields they expect, require, and produce, as well as their corresponding irreps.
    """

    def _init_irreps(
            self,
            irreps_in=None,
            my_irreps_in=None,
            required_irreps_in=None,
            irreps_out=None,
    ):
        """Setup the expected data fields and their irreps for this graph module.

        ``None`` is a valid irreps in the context for anything that is invariant but not well
        described by an ``e3nn.o3.Irreps``. An example are edge indexes in a graph,
         which are invariant but are integers, not ``0e`` scalars.

        Args:
            irreps_in (dict): maps names of all input fields from previous modules or
                data to their corresponding irreps
            my_irreps_in (dict): maps names of fields to the irreps they must have for
                this graph module. Will be checked for consistancy with ``irreps_in``
            required_irreps_in: sequence of names of fields that must be present in
                ``irreps_in``, but that can have any irreps.
            irreps_out (dict): mapping names of fields that are modified/output by
                this graph module to their irreps.
        """
        # Coerce
        if irreps_out is None:
            irreps_out = {}
        if required_irreps_in is None:
            required_irreps_in = []
        if my_irreps_in is None:
            my_irreps_in = {}
        if irreps_in is None:
            irreps_in = {}

        irreps_in = AtomicDataDict._fix_irreps_dict(irreps_in)
        my_irreps_in = AtomicDataDict._fix_irreps_dict(my_irreps_in)
        irreps_out = AtomicDataDict._fix_irreps_dict(irreps_out)

        # positions are *always* 1o, and always present
        if AtomicDataDict.POSITIONS_KEY in irreps_in:
            if irreps_in[AtomicDataDict.POSITIONS_KEY] != o3.Irreps("1x1o"):
                raise ValueError(
                    f"Positions must have irreps 1o, got instead `{irreps_in[AtomicDataDict.POSITIONS_KEY]}`")
        irreps_in[AtomicDataDict.POSITIONS_KEY] = o3.Irreps("1o")

        # edges are also always present
        if AtomicDataDict.EDGE_INDEX_KEY in irreps_in:
            if irreps_in[AtomicDataDict.EDGE_INDEX_KEY] is not None:
                raise ValueError(
                    f"Edge indexes must have irreps None, got instead `{irreps_in[AtomicDataDict.EDGE_INDEX_KEY]}`")
        irreps_in[AtomicDataDict.EDGE_INDEX_KEY] = None

        # Confirm compatibility:
        # with my_irreps_in
        for k in my_irreps_in:
            if k in irreps_in and irreps_in[k] != my_irreps_in[k]:
                raise ValueError(
                    f"The given input irreps {irreps_in[k]} for in_field '{k}' is incompatible with this "
                    f"configuration {type(self)}; should have been {my_irreps_in[k]}")

        # with required_irreps_in
        for k in required_irreps_in:
            if k not in irreps_in:
                raise ValueError(
                    f"This {type(self)} requires in_field '{k}' is in irreps_in")

        # Save stuff
        self.irreps_in = irreps_in

        # The output irreps of any graph module are whatever inputs it has, overwritten with whatever outputs it has.
        new_out = irreps_in.copy()
        new_out.update(irreps_out)
        self.irreps_out = new_out

    def _add_independent_irreps(self, irreps: Dict[str, Any]):
        """
        Insert some independent irreps that need to be exposed to the self.irreps_in and self.irreps_out.
        The terms that have already appeared in the irreps_in will be removed.

        Args:
            irreps (dict): maps names of all new fields
        """

        irreps = {key: irrep for key, irrep in irreps.items() if key not in self.irreps_in}
        irreps_in = AtomicDataDict._fix_irreps_dict(irreps)

        irreps_out = {key: irrep for key, irrep in irreps.items() if key not in self.irreps_out}
        irreps_out = AtomicDataDict._fix_irreps_dict(irreps_out)

        self.irreps_in.update(irreps_in)
        self.irreps_out.update(irreps_out)

    def _make_tracing_inputs(self, n):
        # We impliment this to be able to trace graph modules
        out = []
        for _ in range(n):
            batch = random.randint(1, 4)
            out.append(
                {
                    "forward": ({k: i.randn(batch, -1) for k, i in self.irreps_in.items() if i is not None},)
                }
            )
        return out


class SequentialGraphNetwork(GraphModuleMixin, torch.nn.Sequential):
    r"""A ``torch.nn.Sequential`` of ``GraphModuleMixin``s.

    Args:
        modules (list or dict of ``GraphModuleMixin``s): the sequence of graph modules. If a list, the modules will be named ``"module0", "module1", ...``.
    """

    def __init__(self, modules: Union[Sequence[GraphModuleMixin], Dict[str, GraphModuleMixin]], ):
        if isinstance(modules, dict):
            module_list = list(modules.values())
        else:
            module_list = list(modules)
        # check in/out irreps compatible
        for m1, m2 in zip(module_list, module_list[1:]):
            assert AtomicDataDict._irreps_compatible(m1.irreps_out, m2.irreps_in), \
                f"Incompatible irreps_out from {type(m1).__name__} for " \
                f"input to {type(m2).__name__}: {m1.irreps_out} -> {m2.irreps_in}"

        self._init_irreps(
            irreps_in=module_list[0].irreps_in,
            my_irreps_in=module_list[0].irreps_in,
            irreps_out=module_list[-1].irreps_out,
        )
        # torch.nn.Sequential will name children correctly if passed an OrderedDict
        if isinstance(modules, dict):
            modules = OrderedDict(modules)
        else:
            modules = OrderedDict((f"module{i}", m) for i, m in enumerate(module_list))
        super().__init__(modules)

    @classmethod
    def from_parameters(
            cls,
            shared_params: dict,
            layers: Dict[str, Union[Callable, Tuple[Callable, Dict[str, Any]]]],
            irreps_in: Optional[dict] = None,
    ):
        r"""Construct a ``SequentialGraphModule`` of modules built from a shared set of parameters.

        For some layer, a parameter with name ``param`` will be taken, in order of priority, from:
          1. The specific value in the parameter dictionary for that layer, if provided
          2. ``name_param`` in ``shared_params`` where ``name`` is the name of the layer
          3. ``param`` in ``shared_params``

        Args:
            shared_params (dict): shared parameters from which to pull when instantiating the module
            layers (dict): dictionary mapping unique names of layers to either:
                  1. A callable (such as a class or function) that can be used to ``instantiate`` a module for that layer
                  2. A tuple of such a callable and a dictionary mapping parameter names to values.
                  The given dictionary of parameters will override for this layer values found in ``shared_params``.
                Options 1. and 2. can be mixed.
            irreps_in (optional dict): ``irreps_in`` for the first module in the sequence.

        Returns:
            The constructed SequentialGraphNetwork.
        """
        # note that dictionary ordered gueranteed in >=3.7, so its fine to do an ordered sequential as a dict.
        built_modules = []
        for name, builder in layers.items():
            if not isinstance(name, str):
                raise ValueError(f"`'name'` must be a str; got `{name}`")
            if isinstance(builder, tuple):
                builder, params = builder
            else:
                params = {}
            if not callable(builder):
                raise TypeError(f"The builder has to be a class or a function. got {type(builder)}")

            irreps_in = (built_modules[-1].irreps_out if len(built_modules) > 0 else irreps_in)

            instance, _ = instantiate(builder=builder, prefix=name, positional_args=(dict(irreps_in)),
                                      optional_args=params, all_args=shared_params, )

            if not isinstance(instance, GraphModuleMixin):
                raise TypeError(f"Builder `{builder}` for layer with name `{name}` did not "
                                f"return a GraphModuleMixin, instead got a {type(instance).__name__}")

            built_modules.append(instance)

        return cls(OrderedDict(zip(layers.keys(), built_modules)), )

    def append(self, name: str, module: GraphModuleMixin) -> None:
        r"""Append a module to the SequentialGraphNetwork.

        Args:
            name (str): the name for the module
            module (GraphModuleMixin): the module to append
        """
        assert AtomicDataDict._irreps_compatible(self.irreps_out, module.irreps_in)
        self.add_module(name, module)
        self.irreps_out = dict(module.irreps_out)

    def append_from_parameters(self, shared_params: Dict, name: str, builder: Callable, params=None, ) -> None:
        r"""Build a module from parameters and append it.

        Args:
            shared_params (dict-like): shared parameters from which to pull when instantiating the module
            name (str): the name for the module
            builder (callable): a class or function to build a module
            params (dict, optional): extra specific parameters for this module that take priority
            over those in ``shared_params``
        """
        if params is None:
            params = {}
        instance, _ = instantiate(builder=builder, prefix=name, positional_args=(dict(irreps_in=self[-1].irreps_out)),
                                  optional_args=params, all_args=shared_params, )
        self.append(name, instance)
        return

    def insert(
            self,
            name: str,
            module: GraphModuleMixin,
            after: Optional[str] = None,
            before: Optional[str] = None,
    ) -> None:
        """Insert a module after the module with name ``after``.

        Args:
            name: the name of the module to insert
            module: the moldule to insert
            after: the module to insert after
            before: the module to insert before
        """

        if (before is None) is (after is None):
            raise ValueError("Only one of before or after argument needs to be defined")
        elif before is None:
            insert_location = after
        else:
            insert_location = before

        # This checks names, etc.
        self.add_module(name, module)
        # Now insert in the right place by overwriting
        names = list(self._modules.keys())
        modules = list(self._modules.values())
        idx = names.index(insert_location)
        if before is None:
            idx += 1

        names.insert(idx, name)
        modules.insert(idx, module)

        self._modules = OrderedDict(zip(names, modules))

        module_list = list(self._modules.values())

        # sanity check the compatibility
        if idx > 0:
            assert AtomicDataDict._irreps_compatible(
                module_list[idx - 1].irreps_out, module.irreps_in)

        if len(module_list) > idx:
            assert AtomicDataDict._irreps_compatible(
                module_list[idx + 1].irreps_in, module.irreps_out)

        # insert the new irreps_out to the later modules
        for module_id, next_module in enumerate(module_list[idx + 1:]):
            next_module._add_independent_irreps(module.irreps_out)

        # update the final wrapper irreps_out
        self.irreps_out = dict(module_list[-1].irreps_out)

        return

    def insert_from_parameters(
            self,
            shared_params: Dict,
            name: str,
            builder: Callable,
            params=None,
            after: Optional[str] = None,
            before: Optional[str] = None,
    ) -> None:
        r"""Build a module from parameters and insert it after ``after``.

        Args:
            shared_params (dict-like): shared parameters from which to pull when instantiating the module
            name (str): the name for the module
            builder (callable): a class or function to build a module
            params (dict, optional): extra specific parameters for this module that take priority over those in ``shared_params``
            after: the name of the module to insert after
            before: the name of the module to insert before
        """
        if params is None:
            params = {}
        if (before is None) is (after is None):
            raise ValueError("Only one of before or after argument needs to be defined")
        elif before is None:
            insert_location = after
        else:
            insert_location = before
        idx = list(self._modules.keys()).index(insert_location) - 1
        if before is None:
            idx += 1
        instance, _ = instantiate(
            builder=builder,
            prefix=name,
            positional_args=(dict(irreps_in=self[idx].irreps_out)),
            optional_args=params,
            all_args=shared_params,
        )
        self.insert(after=after, before=before, name=name, module=instance)
        return

    # Copied from https://pytorch.org/docs/stable/_modules/torch/nn/modules/container.html#Sequential
    # with type annotations added
    def forward(self, input: ST) -> ST:
        for module in self:
            input = module(input)
        return input



import torch
from torch.nn import functional as F


def mlm(y_preds: tuple, batch_ys: tuple):
    ls = []
    for i, j in zip(y_preds, batch_ys):
        ls.append(F.l1_loss(i, j, reduction="mean"))
    return sum(ls)


def mlm_alpha(y_preds: tuple, batch_ys: tuple, alphas=None):
    if alphas is None:
        alphas = torch.ones(len(y_preds))
    ls = []
    for i, j, k in zip(y_preds, batch_ys, alphas):
        ls.append(k * F.l1_loss(i, j, reduction="mean"))
    return sum(ls)



# -*- coding: utf-8 -*-

# @Time     : 2021/8/24 13:57
# @Software : PyCharm
# @License  : GNU General Public License v3.0
# @Author   : xxx

#####################################
import os
import warnings
try:
    from pymatgen import SETTINGS
except ImportError:
    from pymatgen.core import SETTINGS


from pymatgen.io.vasp import Potcar, Poscar


def check_PMG_VASP_PSP_DIR():
    d = SETTINGS.get("PMG_VASP_PSP_DIR")
    if d is None:
        warnings.warn("No POTCAR found. Please set the PMG_VASP_PSP_DIR environment in .pmgrc.yaml "
                      "by `set_PMG_VASP_PSP_DIR function` or manually", UserWarning)
        return False
    else:
        return True


def set_PMG_VASP_PSP_DIR(abspath):
    """
    # 初始 potcar,位置为 EXTRACTED_VASP_POTCAR （初始 poscar 数据源在华为盘，百度盘，240）

    # 1.`pmg config - p <EXTRACTED_VASP_POTCAR>  ../../poscar_pmg  ` 产生 poscar_pmg 文件夹

    如果直接有 poscar_pmg 文件夹，(1)可使用此函数设置 poscar_pmg 路径，(2)或者可以直接用第2步

    # 2.`pmg config --add PMG_VASP_PSP_DIR <MY_PSP> ` 设置 poscar_pmg 文件夹

    """
    d = SETTINGS.get("PMG_VASP_PSP_DIR")
    SETTINGS_FILE = os.path.join(os.path.expanduser("~"), ".pmgrc.yaml")
    if d is None:
        if os.path.isfile(SETTINGS_FILE):
            with open(SETTINGS_FILE, "a+") as f:
                f.write("\nPMG_VASP_PSP_DIR: {}}".format(abspath))
    else:
        if os.path.isfile(SETTINGS_FILE):
            with open(SETTINGS_FILE, "r") as f:
                lines = f.readlines()
            lines = [i for i in lines if "VASP_PSP_DIR" not in i]
            with open(SETTINGS_FILE, "w+") as f:
                f.writelines(lines)
                f.write("\nPMG_VASP_PSP_DIR: {}".format(abspath))
        print("PMG_VASP_PSP_DIR is refreshed to `{}`".format(abspath))


class Potcar2(Potcar):

    def __init__(self, symbols=None, functional=None, sym_potcar_map=None):
        if sym_potcar_map is None:
            assert check_PMG_VASP_PSP_DIR()
        super().__init__(symbols=symbols, functional=functional, sym_potcar_map=sym_potcar_map)

    @classmethod
    def from_poscar(cls, poscar: Poscar, functional=None, sym_potcar_map=None):
        symbol = poscar.site_symbols
        return cls(symbols=symbol, functional=functional, sym_potcar_map=sym_potcar_map)


"""
This part sparse db (ase) data.

Examples:
    >>> dict_data = sparse_ase(db="data.db")

Examples:
    >>> from hpc2ml.data.structuretodata import StructureToData
    >>> addsap = StructureToData()
    >>> source_path = "."
    >>> res = addsap.sparse_source_data(source_file="data.db", source_path=source_path, fmt="ase",)
"""

from typing import Dict, Any, Union

from ase.db.core import reserved_keys, connect
from ase.db.jsondb import JSONDatabase
from ase.db.row import AtomsRow
from ase.db.sqlite import SQLite3Database

from hpc2ml.data.io.ase import aaa


def db_row2dct(row: AtomsRow
               ) -> Dict[str, Any]:
    """Convert row to dict."""
    temp = {}
    temp.update(row.data)
    temp.update(row.key_value_pairs)
    dct = {i: row.__dict__[i] for i in reserved_keys if i in row.__dict__}
    temp.update(dct)
    atoms = row.toatoms()
    st = aaa.get_structure(atoms)
    temp.update({"structure": st})
    id_str = temp.get("id")
    return {id_str: temp}


def sparse_ase(db: Union[SQLite3Database, JSONDatabase, str]) -> Dict[str, Any]:
    """Sparse ase data."""
    db = connect(db)
    dct = {}
    for row in db.select():
        dct.update(db_row2dct(row))
    return dct



# -*- coding: utf-8 -*-

# @Time    : 2021/7/16 11:14
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause
from typing import Union, List

import numpy as np
from ase import Atoms
from mgetool.tool import parallelize
from pymatgen.analysis.energy_models import EwaldElectrostaticModel
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor

aaa = AseAtomsAdaptor()


def get_stability(structure: Union[Structure, Atoms]):
    """get energy by EwaldElectrostaticModel."""
    if isinstance(structure, Atoms):
        structure = aaa.get_structure(atoms=structure)
    eem = EwaldElectrostaticModel(acc_factor=8.0)
    try:
        return eem.get_energy(structure)
    except BaseException:
        return 10.0


def filter_stability(structure: List[Union[Structure, Atoms]], n_jobs=1, ):
    energies = parallelize(n_jobs, get_stability, structure, respective=False, tq=True,
                           batch_size='auto', mode="j")
    return np.array(energies)




import warnings
from typing import Dict, Optional, Union, List

import ase.data
import torch

from hpc2ml.data.nequlip_support import AtomicDataDict
from hpc2ml.data.nequlip_support.AtomicDataDict import ST


class TypeMapper:
    """Based on a configuration, map atomic numbers to types."""

    num_types: int
    chemical_symbol_to_type: Optional[Dict[str, int]]
    type_names: List[str]
    _min_Z: int

    def __init__(
            self,
            type_names: Optional[List[str]] = None,
            chemical_symbol_to_type: Optional[Dict[str, int]] = None,
            chemical_symbols: Optional[List[str]] = None,
    ):
        if chemical_symbols is not None:
            if chemical_symbol_to_type is not None:
                raise ValueError(
                    "Cannot provide both `chemical_symbols` and `chemical_symbol_to_type`"
                )
            # repro old, sane NequIP behaviour
            # checks also for validity of keys
            atomic_nums = [ase.data.atomic_numbers[sym] for sym in chemical_symbols]
            # https://stackoverflow.com/questions/29876580/how-to-sort-a-list-according-to-another-list-python
            chemical_symbols = [
                e[1] for e in sorted(zip(atomic_nums, chemical_symbols))
            ]
            chemical_symbol_to_type = {k: i for i, k in enumerate(chemical_symbols)}
            del chemical_symbols

        # Build from chem->type mapping, if provided
        self.chemical_symbol_to_type = chemical_symbol_to_type
        if self.chemical_symbol_to_type is not None:
            # Validate
            for sym, type in self.chemical_symbol_to_type.items():
                assert sym in ase.data.atomic_numbers, f"Invalid chemical symbol {sym}"
                assert 0 <= type, f"Invalid type number {type}"
            assert set(self.chemical_symbol_to_type.values()) == set(
                range(len(self.chemical_symbol_to_type))
            )
            if type_names is None:
                # Make type_names
                type_names = [None] * len(self.chemical_symbol_to_type)
                for sym, type in self.chemical_symbol_to_type.items():
                    type_names[type] = sym
            else:
                # Make sure they agree on types
                # We already checked that chem->type is contiguous,
                # so enough to check length since type_names is a list
                assert len(type_names) == len(self.chemical_symbol_to_type)
            # Make mapper array
            valid_atomic_numbers = [
                ase.data.atomic_numbers[sym] for sym in self.chemical_symbol_to_type
            ]
            self._min_Z = min(valid_atomic_numbers)
            self._max_Z = max(valid_atomic_numbers)
            Z_to_index = torch.full(
                size=(1 + self._max_Z - self._min_Z,), fill_value=-1, dtype=torch.long
            )
            for sym, type in self.chemical_symbol_to_type.items():
                Z_to_index[ase.data.atomic_numbers[sym] - self._min_Z] = type
            self._Z_to_index = Z_to_index
            self._index_to_Z = torch.zeros(
                size=(len(self.chemical_symbol_to_type),), dtype=torch.long
            )
            for sym, type_idx in self.chemical_symbol_to_type.items():
                self._index_to_Z[type_idx] = ase.data.atomic_numbers[sym]
            self._valid_set = set(valid_atomic_numbers)
        # check
        if type_names is None:
            raise ValueError(
                "None of chemical_symbols, chemical_symbol_to_type, nor type_names was provided; exactly one is required"
            )
        # validate type names
        assert all(
            n.isalnum() for n in type_names
        ), "Type names must contain only alphanumeric characters"
        # Set to however many maps specified -- we already checked contiguous
        self.num_types = len(type_names)
        # Check type_names
        self.type_names = type_names

    def __call__(
            self, data: Union[ST], types_required: bool = True
    ) -> Union[ST]:
        if AtomicDataDict.ATOM_TYPE_KEY in data:
            if AtomicDataDict.ATOMIC_NUMBERS_KEY in data:
                warnings.warn(
                    "Data contained both ATOM_TYPE_KEY and ATOMIC_NUMBERS_KEY; ignoring ATOMIC_NUMBERS_KEY"
                )
        elif AtomicDataDict.ATOMIC_NUMBERS_KEY in data:
            assert (
                    self.chemical_symbol_to_type is not None
            ), "Atomic numbers provided but there is no chemical_symbols/chemical_symbol_to_type mapping!"
            atomic_numbers = data[AtomicDataDict.ATOMIC_NUMBERS_KEY]
            del data[AtomicDataDict.ATOMIC_NUMBERS_KEY]

            data[AtomicDataDict.ATOM_TYPE_KEY] = self.transform(atomic_numbers)
        else:
            if types_required:
                raise KeyError(
                    "Data doesn't contain any atom type information (ATOM_TYPE_KEY or ATOMIC_NUMBERS_KEY)"
                )
        return data

    def transform(self, atomic_numbers):
        """core function to transform an array to specie index list"""

        if atomic_numbers.min() < self._min_Z or atomic_numbers.max() > self._max_Z:
            bad_set = set(torch.unique(atomic_numbers).cpu().tolist()) - self._valid_set
            raise ValueError(
                f"Data included atomic numbers {bad_set} that are not part of the atomic number -> type mapping!"
            )

        return self._Z_to_index.to(device=atomic_numbers.device)[
            atomic_numbers - self._min_Z
            ]

    def untransform(self, atom_types):
        """Transform atom types back into atomic numbers"""
        return self._index_to_Z[atom_types].to(device=atom_types.device)

    @property
    def has_chemical_symbols(self) -> bool:
        return self.chemical_symbol_to_type is not None

    @staticmethod
    def format(
            data: list, type_names: List[str], element_formatter: str = ".6f"
    ) -> str:
        data = torch.as_tensor(data) if data is not None else None
        if data is None:
            return f"[{', '.join(type_names)}: None]"
        elif data.ndim == 0:
            return (f"[{', '.join(type_names)}: {{:{element_formatter}}}]").format(data)
        elif data.ndim == 1 and len(data) == len(type_names):
            return (
                    "["
                    + ", ".join(
                f"{{{i}[0]}}: {{{i}[1]:{element_formatter}}}"
                for i in range(len(data))
            )
                    + "]"
            ).format(*zip(type_names, data))
        else:
            raise ValueError(
                f"Don't know how to format data=`{data}` for types {type_names} with element_formatter=`{element_formatter}`"
            )



"""
From ocp
"""

import torch
import torch.nn.functional as F


class Act(torch.nn.Module):
    def __init__(self, act, slope=0.05):
        super(Act, self).__init__()
        self.act = act
        self.slope = slope
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, input):
        if self.act == "relu":
            return F.relu(input)
        elif self.act == "leaky_relu":
            return F.leaky_relu(input)
        elif self.act == "sp":
            return F.softplus(input, beta=1)
        elif self.act == "leaky_sp":
            return F.softplus(input, beta=1) - self.slope * F.relu(-input)
        elif self.act == "elu":
            return F.elu(input, alpha=1)
        elif self.act == "leaky_elu":
            return F.elu(input, alpha=1) - self.slope * F.relu(-input)
        elif self.act == "ssp":
            return F.softplus(input, beta=1) - self.shift
        elif self.act == "leaky_ssp":
            return (
                    F.softplus(input, beta=1)
                    - self.slope * F.relu(-input)
                    - self.shift
            )
        elif self.act == "tanh":
            return torch.tanh(input)
        elif self.act == "leaky_tanh":
            return torch.tanh(input) + self.slope * input
        elif self.act == "swish":
            return torch.sigmoid(input) * input
        else:
            raise RuntimeError(f"Undefined activation called {self.act}")



# -*- coding: utf-8 -*-

# @Time  : 2023/4/19 17:23
# @Author : boliqq07
# @Software: PyCharm
# @License: MIT License
import ase
import numpy as np
from ase import Atom, Atoms
from ase.build import molecule
from ase.visualize import view
from numpy import random
from numpy.linalg import norm


class SubstituteWithRotatedB:
    """Substitute atoms in atoms.

    such as: ABX3 -> ACX3

    such as: ABX3 -> ABO3

    such as: ABX3 -> (CH3NH3)BX3

    """

    def __init__(self, strategy="same", **kwargs):
        assert strategy in ["same", "mirror", "random", "auto"]

        if strategy == "auto":
            assert "space_group_number" in kwargs

        self.strategy = strategy
        self.kwargs = {"space_group_number":1}
        self.kwargs.update(kwargs)

    def sub_atoms(self, atoms: ase.Atoms, a: str, b: [str, ase.Atom, ase.Atoms], index=None):
        """Substitute atom/atoms."""
        atoms = atoms.copy()
        if a not in atoms.symbols.species():
            raise TypeError("Cant find the a element {}".format(a))

        indexs = atoms.symbols.indices()[a]

        if isinstance(index, (list, tuple)):
            try:
                indexs = indexs[tuple(index)]
            except IndexError:
                raise TypeError("Number of index:{} is out of number of a atoms {}".format(len(index), len(indexs)))

        if isinstance(b, ase.Atoms):
            if len(b) >= 2:
                bbs = self.get_rotate_atoms_list(b, atoms=atoms, index=indexs)
                for indi, bb in zip(indexs, bbs):
                    atoms.extend(bb)
                del atoms[indexs]
            else:
                atoms.symbols[indexs] = b

        elif isinstance(b, str) or isinstance(b, Atom):
            atoms.symbols[indexs] = b
        else:
            raise TypeError("b is str element name or ase.Atoms")

        return atoms

    def get_rotate_atoms_list(self, b, atoms, index=None):
        """rotate with strategy."""
        b = b.copy()

        if self.strategy == "same":
            bs = self._same(b, atoms=atoms, index=index, **self.kwargs)
        elif self.strategy == "auto":
            bs = self._auto(b, atoms=atoms, index=index, **self.kwargs)
        elif self.strategy == "random":
            bs = self._random(b, atoms=atoms, index=index, **self.kwargs)
        elif self.strategy == "mirror":
            bs = self._mirror(b, atoms=atoms, index=index, **self.kwargs)
        else:
            raise NotImplementedError

        bs = self._process(bs, atoms=atoms, index=index)
        return bs

    @staticmethod
    def _process(bs, atoms, index=None):
        """set the position for each b atoms"""
        for indi, b in zip(index, bs):
            b.center()
            b.set_pbc(False)
            b.set_positions(atoms.get_positions()[indi] + b.get_positions())
        return bs

    @staticmethod
    def _random(b: [str, ase.Atom, ase.Atoms], atoms=None, index=None, random_state=None, **kwargs):

        """Substitute atom/atoms in random direction."""

        random.seed(random_state)
        bs = []
        for _ in index:
            bb = b.copy()
            angle = random.randint(1, 360)
            axi = random.random(size=3)
            bb.rotate(a=angle, v=axi)
            bs.append(bb)
        return bs

    @staticmethod
    def _same(b: [str, ase.Atom, ase.Atoms], atoms=None, index=None, angle=0, v="x", **kwargs):

        """Substitute atom/atoms in same direction."""
        bs = []
        for _ in index:
            bb = b.copy()
            bb.rotate(a=angle, v=v)
            bs.append(bb)
        return bs

    @staticmethod
    def _nothing(b: [str, ase.Atom, ase.Atoms], atoms=None, index=None, **kwargs):

        """do nothing."""
        bs = []
        for _ in index:
            bb = b.copy()
            bs.append(bb)
        return bs


    def _auto(self, b: [str, ase.Atom, ase.Atoms], atoms: Atoms = None, index=None, **kwargs):

        """judge auto by space_group_number."""
        space_group_number = kwargs["space_group_number"]
        if space_group_number == 221:
            return self._nothing(b, atoms=atoms, index=index, **kwargs)

        elif space_group_number in [62, 127]:
            try:
                abc_ = atoms.cell.cellpar()[:3]
                c_ = sum(max(abc_) - abc_) - sum(min(abc_) - abc_)
                if c_ >= 0:
                    axi = np.argmax(abc_)
                else:
                    axi = np.argmin(abc_)
            except TypeError:
                axi = 2

            positions = atoms.get_scaled_positions()[index]

            tol = kwargs["tol"] if "tol" in kwargs else 0.1

            index_group = _merge(positions[:, axi], tol=tol)
            index_groupd = [j for i, j in enumerate(index_group) if i % 2 == 0]
            index_groupd = [j for i in index_groupd for j in i]

            if "v" not in kwargs:
                v = {2: "z", 1: "y", 0: "x"}[axi]
            else:
                v = kwargs["v"]

            bs = []
            for i in range(len(index)):
                bb = b.copy()

                if i in index_groupd:
                    bb.rotate(a=180, v=v)
                if i % 2 == 1:
                    bb.rotate(a=90, v=(1, 1, 0))
                bs.append(bb)
            return bs
        else:
            return self._nothing(b, atoms=atoms, index=index, **kwargs)

    def _mirror(self, b: [str, ase.Atom, ase.Atoms], atoms: Atoms, index: np.ndarray, angle=180, v="z", **kwargs):

        """Substitute atom/atoms in same direction."""

        if len(index) == 1:
            return self._nothing(b, atoms=atoms, index=index, **kwargs)

        positions = atoms.get_scaled_positions()[index]

        distance = norm(positions, axis=1)

        tol = kwargs["tol"] if "tol" in kwargs else 0.1

        index_group = _merge(distance, tol=tol)
        index_groupd = [j for i, j in enumerate(index_group) if i%2==0]
        index_groupd = [j for i in index_groupd for j in i]

        bs = []
        for i in range(len(index)):
            bb = b.copy()
            if i in index_groupd:
                bb.rotate(a=angle, v=v)
            bs.append(bb)
        return bs


def _merge(distance, tol=0.1):
    distance_index = np.argsort(distance)
    distance = np.sort(distance)
    i = 0
    group = [[0, ]]
    while i < len(distance) - 1:
        if abs(distance[i] - distance[i + 1]) < tol:
            group[-1].append(i + 1)
        else:
            group.append([i + 1])
        i += 1

    index_group = [[distance_index[ii] for ii in i] for i in group]

    return index_group


if __name__ == "__main__":
    from perovskite.core.general_standard import perovskite_prim_221, perovskite_conv_221, perovskite_prim_62, \
        perovskite_conv_62
    atomss = perovskite_conv_62(a_atom="Ba", b_atom="Ti", c_atom="O", a=7, size=(1, 1, 1))
    # atomss = perovskite_conv_221(a_atom="Ba", b_atom="Ti", c_atom="O", a=7, size=(1, 1, 1))
    no = molecule('HCF3')
    no.get_scaled_positions()
    view(atomss)
    att =  SubstituteWithRotatedB(strategy="auto",space_group_number=62)
    # att =  BaseSubstitute(strategy="mirror",angle=90, v="x")
    atoms =att.sub_atoms(atomss, "Ba", no, index=None)
    view(atoms)


# -*- coding: utf-8 -*-

# @Time    : 2021/7/18 23:54
# @Email   : 986798607@qq.com
# @Software: PyCharm
# @License: BSD 3-Clause

import numpy as np

from ase import Atoms
from pymatgen.core import Element


def r_spheres_effective(atoms: Atoms):
    """
    For A site.
    This function just calculated with the ionic radii of pymatgen with  universe valence.
    but for some mulecule, the  valence are changed. Thus , For reference only.
    r_eff= rmass + rion, where rmass
        is defined as the distance between the centre of mass of the A‑site organic cation and the
        atom with the largest distance to the centre of mass (excluding hydrogen atoms), and rion
        is the corresponding ionic radius of the aforementioned atom. rXeff can be defined in a similar
        way to rAeff."""

    com = atoms.get_center_of_mass(scaled=False)
    atoms.set_center_of_mass(com, scaled=False)

    posi = atoms.positions
    distance = np.linalg.norm(posi, axis=1)
    drop_H = [i for i in range(len(atoms)) if atoms[i].number == 1]

    left = []
    for i in range(len(atoms)):
        if i in drop_H:
            pass
        else:
            left.append(i)

    left = np.array(left)

    distance = distance[left]
    index = np.argmax(distance)
    r_mass = np.max(distance)
    z = atoms[left[index]].number
    if z == 1:
        r_ion = 0.0
    elif z == 6:  # for C element, the C-4 use 177pm.
        r_ion = 1.77
    else:
        ele = Element.from_Z(z)
        r_ion_dict = ele.ionic_radii
        r_ion = list(r_ion_dict.values())[0]
    return r_mass + r_ion


def r_cylinders_effective(atoms: Atoms):
    """
    For C site.

    This function just calculated with the ionic radii of pymatgen with  universe valence.
    but for some mulecule, the  valence are changed. Thus , For reference only.

    return r_eff,r_h
    """

    com = atoms.get_center_of_mass(scaled=False)
    atoms.set_center_of_mass(com, scaled=False)

    drop_H = [i for i in range(len(atoms)) if atoms[i].number == 1]

    left = []
    for i in range(len(atoms)):
        if i in drop_H:
            pass
        else:
            left.append(i)
    left = np.array(left)

    positions = atoms.positions[:, np.newaxis, :]
    positions2 = atoms.positions[np.newaxis, :, :]
    posi = positions - positions2

    distance = np.linalg.norm(posi, axis=2)
    distance = distance[left][left]
    index = np.argmax(distance, axis=0)

    index = [np.argmax(index), max(index)]

    r_h = np.max(distance)

    for indexi in index:
        z = atoms[left[indexi]].number
        if z == 1:
            r_ion = 0.0
        elif z == 6:  # for C element, the C-4 use 177pm.
            r_ion = 1.77
        else:
            ele = Element.from_Z(z)
            r_ion_dict = ele.ionic_radii
            r_ion = list(r_ion_dict.values())[0]
        r_h += r_ion

    # part2

    com = atoms.get_center_of_mass(scaled=False)
    atoms.set_center_of_mass(com, scaled=False)

    posi = atoms.positions
    distance2 = np.linalg.norm(posi, axis=1)

    distance2 = distance2[left]
    index2 = np.argmax(distance2)

    orita = posi[left[index]]
    orita = orita[0] - orita[1]

    vec = posi[left[index2]]

    norm_orita = np.linalg.norm(orita)
    norm_vec = np.linalg.norm(vec)
    cos_angle = orita.dot(vec) / (norm_orita * norm_vec)
    sin_angle = (1 - cos_angle ** 2) ** 0.5
    r_mass = sin_angle * norm_vec

    z = atoms[left[index2]].number

    if z == 1:
        r_ion = 0.0
    elif z == 6:  # for C element, the C-4 use 177pm.
        r_ion = 1.77
    else:
        ele = Element.from_Z(z)
        r_ion_dict = ele.ionic_radii
        r_ion = list(r_ion_dict.values())[0]

    r_eff = r_mass + r_ion

    return r_eff, r_h


def get_all_ion_radii():
    for i in range(1, 100):
        ele = Element.from_Z(i)
        r_ion_dict = ele.ionic_radii
        r_ion = list(r_ion_dict.values())
        if len(r_ion) > 0:
            r_ion = r_ion[0]
        else:
            r_ion = np.NaN

        print(ele.symbol, r_ion)


if __name__ == "__main__":
    get_all_ion_radii()



from typing import Dict, Any

_SPECIAL_IRREPS = [None]

from e3nn import o3


def _fix_irreps_dict(d: Dict[str, Any]):

    return {k: (i if i in _SPECIAL_IRREPS else o3.Irreps(i)) for k, i in d.items()}


def _irreps_compatible(ir1: Dict[str, o3.Irreps], ir2: Dict[str, o3.Irreps]):

    return all(ir1[k] == ir2[k] for k in ir1 if k in ir2)



import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import ReLU, GRU
from torch_geometric.nn import MessagePassing, aggr
from torch_scatter import scatter

from hpc2ml.data.embedding.cohesive_energy import atomic_energy
from hpc2ml.function.support import add_edge_total
from hpc2ml.nn.activations import Act
from hpc2ml.nn.forces import Forces
from hpc2ml.nn.lj import LJ
from hpc2ml.nn.morse import Morse
from hpc2ml.nn.stress import Stress


class PotConv(MessagePassing):
    """Potential network.

    pot="morse" : Morse potential

    pot="lj": Lennard-Jones potential

    Keep edge_attr is the raw distance_vec, edge_weight is the raw care_ distance.
    """

    def __init__(self, in_channels, out_channels, nc_edge_hidden, pot="morse", mode="xyz", batch_norm=True):
        super().__init__(aggr='mean', flow="source_to_target")
        if pot == "morse":
            self.potlayer = Morse()
        elif pot == "lj":
            self.potlayer = LJ()
        else:
            raise NotImplementedError

        self.mode = mode

        if self.mode == "xyz":
            assert nc_edge_hidden == 3  # just for x,y,z vector.
        else:
            assert nc_edge_hidden == 1  # just for r vector.

        self.lin1 = nn.Linear(2 * in_channels + nc_edge_hidden + 1, 2 * in_channels, bias=False)
        self.lin1p = nn.Linear(2 * in_channels + 3, 2 * out_channels)  # not used perhaps
        self.lin2 = nn.Linear(2 * in_channels, 2 * out_channels)
        self.lin3 = nn.Linear(2 * out_channels, out_channels)

        self.act1_1p = Act("ssp")
        self.act2_3 = Act("ssp")

        self.batch_norm = batch_norm

        if batch_norm and self.mode == "xyz":
            self.bm = nn.BatchNorm1d(nc_edge_hidden)
        else:
            self.bm = nn.BatchNorm1d(1)

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.lin1.weight)
        nn.init.xavier_uniform_(self.lin2.weight)
        nn.init.xavier_uniform_(self.lin3.weight)

    def forward(self, x, edge_index, edge_weight, edge_attr, z):

        """keep edge_attr is the raw distance_vec, edge_weight is the raw care_ distance"""

        x = self.propagate(edge_index, x=x, edge_weight=edge_weight, edge_attr=edge_attr, z=z)
        x = self.lin2(x)
        x = self.act2_3(x)
        x = self.lin3(x)

        return x

    def message(self, z_i, z_j, x_i, x_j, edge_weight, edge_attr):
        """
        The input edge_index is the neighbor_index and center_index.

        1. If used edge_index (simple torch tensor object),
            Dut to the MessagePassing using flow='source_to_target' default. (i=1 and j=0)

            The edge_index_i=center_index,  edge_index_j=neighbor_index finally.

        2. If used adj_t (SparseTensor),

            In default,

            adj_t = SparseTensor(
                    row=store.edge_index[1], col=store.edge_index[0],
                    value=None if self.attr is None or self.attr not in store else
                    store[self.attr], sparse_sizes=store.size()[::-1],
                    is_sorted=True, trust_data=True)

            Finally, the edge_index_i=adj_t.storage.row()=center_index, edge_index_j=adj_t.storage.col()=neighbor_index,
            in MessagePassing.

        3.  If used edge_index (torch, and edge_index.is_sparse), (Not suggested !!!).
            edge_index_i = edge_index.indices()[0]
            edge_index_j = edge_index.indices()[1]
            please make sure the rank of center_index and neighbor_index in edge_index when you generate it!!!

        The rank is same in x (x_i and x_j),  z (x_i and x_j)
        """
        if self.mode == "xyz":
            return self.message_xyz(z_i, z_j, x_i, x_j, edge_weight, edge_attr)

        else:
            return self.message_r(z_i, z_j, x_i, x_j, edge_weight, edge_attr)

    def message_r(self, z_i, z_j, x_i, x_j, edge_weight, edge_attr):

        W = self.potlayer(cen=z_i, nei=z_j, r=edge_weight)
        if self.batch_norm:
            W = self.bm(W)
        x = self.lin1(torch.cat([x_i, x_j, edge_weight.view(-1, 1), edge_attr], dim=1))
        x = x * W
        return x

    def message_xyz(self, z_i, z_j, x_i, x_j, edge_weight, edge_attr):

        W = self.potlayer(cen=z_i, nei=z_j, r=edge_attr)
        if self.batch_norm:
            W = self.bm(W)
        x = self.lin1(torch.cat([x_i, x_j, edge_weight.view(-1, 1), edge_attr], dim=1))
        x = self.act1_1p(x)
        x = self.lin1p(torch.cat([x, W], dim=1))

        return x


class NNConv(MessagePassing):
    def __init__(self, in_channels, out_channels, nc_edge_hidden, cutoff=5.0, mlp=None):
        super().__init__(aggr='mean', flow="source_to_target")

        if mlp is None:
            self.mlp = nn.Sequential(nn.Linear(nc_edge_hidden, 32), ReLU(), nn.Linear(32, in_channels))
        else:
            self.mlp = mlp

        self.cutoff = cutoff

        self.lin1 = nn.Linear(in_channels, in_channels, bias=False)
        self.lin2 = nn.Linear(in_channels, out_channels)
        self.lin3 = nn.Linear(out_channels, out_channels)
        self.act2_3 = Act("ssp")

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.lin1.weight)
        nn.init.xavier_uniform_(self.lin2.weight)
        nn.init.xavier_uniform_(self.lin3.weight)
        self.lin2.bias.data.fill_(0)
        self.lin3.bias.data.fill_(0)

    def forward(self, x, edge_index, edge_weight, edge_attr, z=None):
        C = torch.cos(edge_weight * torch.pi / self.cutoff) * 0.5 + 0.5
        W = self.mlp(edge_attr) * C.view(-1, 1)

        x = self.lin1(x)
        x = self.propagate(edge_index, x=x, W=W)
        x = self.lin2(x)
        x = self.act2_3(x)
        x = self.lin3(x)

        return x

    def message(self, x_j, W):
        """
        The input edge_index is the neighbor_index and center_index.

        1. If used edge_index (simple torch tensor object),
            Dut to the MessagePassing using flow='source_to_target' default. (i=1 and j=0)

            The edge_index_i=center_index,  edge_index_j=neighbor_index finally.

        2. If used adj_t (SparseTensor),

            In default,

            adj_t = SparseTensor(
                    row=store.edge_index[1], col=store.edge_index[0],
                    value=None if self.attr is None or self.attr not in store else
                    store[self.attr], sparse_sizes=store.size()[::-1],
                    is_sorted=True, trust_data=True)

            Finally, the edge_index_i=adj_t.storage.row()=center_index, edge_index_j=adj_t.storage.col()=neighbor_index,
            in MessagePassing.

        3.  If used edge_index (torch, and edge_index.is_sparse), (Not suggested !!!).
            edge_index_i = edge_index.indices()[0]
            edge_index_j = edge_index.indices()[1]
            please make sure the rank of center_index and neighbor_index in edge_index when you generate it!!!

        The rank is same in x (x_i and x_j).
        """
        return x_j * W


class CGGRUForceStress(torch.nn.Module):
    """
    Note:
        1. direct_force=False and device="cuda:i" is un-compacted due to RNNNet can double backward.
        keep using device = "cpu" if direct_force=False."""

    def __init__(self, nfeat_node=19, nc_edge_hidden=1, dim=64, cutoff=5.0, n_block=3, get_force=True,
                 direct_force=True, get_stress=True, readout="set2set", mode="r", try_add_edge_msg=False,
                 use_pbc=True, pot="morse",
                 **kwargs):
        super().__init__()

        if direct_force is False and get_force is True:
            assert try_add_edge_msg == True  # Use to `_generate_graph` function.
            if nc_edge_hidden != 3:
                warnings.warn("nc_edge_hidden should changed to 3, due `_generate_graph` function just add edge_attr "
                              "with shape (n_bond,3)")
            nc_edge_hidden = 3  # Due to `_generate_graph` function just add edge_attr with shape (n_bond,3)

        self.lin0 = nn.Linear(nfeat_node, dim)
        self.act0 = Act("leaky_relu")

        if pot in ["morse", "lj"]:

            if nc_edge_hidden==1:
                mode ="r"
            elif nc_edge_hidden==3:
                mode ="xyz"
            else:
                raise NotImplementedError
            conv0 = PotConv(in_channels=dim, out_channels=dim, nc_edge_hidden=nc_edge_hidden, pot=pot, mode=mode)
        else:
            conv0 = NNConv(in_channels=dim, out_channels=dim, nc_edge_hidden=nc_edge_hidden, mlp=None, cutoff=cutoff)

        self.gru = GRU(dim, dim)

        assert n_block >= 2

        self.conv_list = nn.ModuleList([conv0, ])
        for i in range(n_block - 1):
            self.conv_list.append(NNConv(in_channels=dim, out_channels=dim,
                                         nc_edge_hidden=nc_edge_hidden,
                                         mlp=None, cutoff=cutoff))

        self.readout = readout

        if self.readout == "set2set":
            self.set2set = aggr.Set2Set(dim, processing_steps=3)
            self.lin1 = torch.nn.Linear(2 * dim, dim)
        else:
            self.lin1 = torch.nn.Linear(dim, dim)

        self.lin2 = torch.nn.Linear(dim, 1)
        self.readout = readout
        self.get_force = get_force
        self.get_stress = get_stress
        self.direct_force = direct_force

        self.cutoff = cutoff
        self.use_pbc = use_pbc
        self.try_add_edge_msg = try_add_edge_msg

        if self.direct_force:
            self.force_layer = Forces(dim)
        else:
            self.try_add_edge_msg = True

        if self.get_stress:
            self.stress_layer = Stress(input_dim=dim)

        energy_ref = torch.from_numpy(atomic_energy).view(-1, 1).float()
        self.register_buffer('energy_ref', energy_ref)

        self.atom_ref = nn.Embedding(100, 1)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        for interaction in self.conv_list:
            interaction.reset_parameters()
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        self.lin1.bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)

    @staticmethod
    def _generate_graph(data, cutoff, use_pbc, max_neighbors=32):
        return add_edge_total(data, cutoff=cutoff, use_pbc=use_pbc, max_neighbors=max_neighbors,
                              use_exist_index=True)

    def forward(self, data):
        if not self.direct_force:
            data.pos.requires_grad_(True)

        if self.try_add_edge_msg:
            data = self._generate_graph(data, self.cutoff, self.use_pbc, max_neighbors=32)

        out = self.act0(self.lin0(data.x))
        h = out.unsqueeze(0)

        for convi in self.conv_list:
            m = F.leaky_relu(convi(out, data.edge_index, data.edge_weight, data.edge_attr, data.z.view(-1, 1)))
            if not self.direct_force:
                with torch.backends.cudnn.flags(enabled=False):
                    out, h = self.gru(m.unsqueeze(0), h)
            else:
                out, h = self.gru(m.unsqueeze(0), h)
            out = out.squeeze(0)

        if self.atom_ref is not None:
            out1 = out + self.atom_ref(data.z.view(-1))
        else:
            out1 = out

        if self.energy_ref is not None:
            out1 = out1 + self.energy_ref[data.z.view(-1)]

        if self.readout == "set2set":
            if not self.direct_force:
                with torch.backends.cudnn.flags(enabled=False):
                    energy = self.set2set(out1, data.batch)

            else:
                energy = self.set2set(out1, data.batch)

        else:
            energy = scatter(out1, data.batch, dim=0, reduce=self.readout)

        energy = F.leaky_relu(self.lin1(energy))
        energy = self.lin2(energy)
        energy = energy.view(-1)

        ########################

        if self.get_force:
            if not self.direct_force:
                forces = -1 * (
                    torch.autograd.grad(energy, data.pos,
                                        grad_outputs=torch.ones_like(energy),
                                        create_graph=True, )
                    [0])
            else:
                forces = self.force_layer(out)

            if self.get_stress:
                stress = self.stress_layer(out, data)

        if self.get_force:
            if self.get_stress:
                return energy, forces, stress
            else:
                return energy, forces
        else:
            return energy



"""
Operation utilities on lists and arrays, for torch,

Notes
    just for network.
"""

from collections import abc
from collections.abc import Mapping, Sequence
from functools import wraps
from typing import Any
from typing import Union, List, Optional

import numpy as np
import torch
from torch import Tensor
from torch.nn import Module
from torch_geometric.data.data import BaseData
from torch_geometric.data.storage import BaseStorage
from torch_sparse import SparseTensor


def to_list(x: Union[abc.Iterable, np.ndarray]) -> List:
    """
    If x is not a list, convert it to list
    """
    if isinstance(x, abc.Iterable):
        return list(x)
    if isinstance(x, np.ndarray):
        return x.tolist()  # noqa
    return [x]


def fast_label_binarize(value: List, labels: List) -> List[int]:
    """Faster version of label binarize

    `label_binarize` from scikit-learn is slow when run 1 label at a time.
    `label_binarize` also is efficient for large numbers of classes, which is not
    common in `megnet`

    Args:
        value: Value to encode
        labels (list): Possible class values
    Returns:
        ([int]): List of integers
    """

    if len(labels) == 2:
        return [int(value == labels[0])]
    output = [0] * len(labels)
    if value in labels:
        output[labels.index(value)] = 1
    return output


def check_shape(array: Optional[np.ndarray], shape: Sequence) -> bool:
    """
    Check if array complies with shape. Shape is a sequence of
    integer that may end with None. If None is at the end of shape,
    then any shapes in array after that dimension will match with shape.

    Example: array with shape [10, 20, 30, 40] matches with [10, 20, None], but
        does not match with shape [10, 20, 30, 20]

    Args:
        array (np.ndarray or None): array to be checked
        shape (Sequence): integer array shape, it may ends with None
    Returns: bool
    """
    if array is None:
        return True
    if all(i is None for i in shape):
        return True

    array_shape = array.shape
    valid_dims = [i for i in shape if i is not None]
    n_for_check = len(valid_dims)
    return all(i == j for i, j in zip(array_shape[:n_for_check], valid_dims))


def getter_arr(obj, pi):
    """Get prop.
    """
    if "." in pi:
        pis = list(pi.split("."))
        pis.reverse()
        while len(pis):
            s = pis.pop()
            obj = getter_arr(obj, s)
        return obj
    elif "()" in pi:
        return getattr(obj, pi[:-2])()

    else:
        return getattr(obj, pi)


def temp_jump(mark=0, temp_device=None, old_device=None, temp=True, back=True):
    def f(func):
        return _temp_jump(func, mark=mark, temp_device=temp_device, old_device=old_device, temp=temp, back=back)

    return f


def temp_jump_cpu(mark=0, temp_device="cpu", old_device=None, temp=True, back=True):
    def f(func):
        return _temp_jump(func, mark=mark, temp_device=temp_device, old_device=old_device, temp=temp, back=back)

    return f


def _temp_jump(func, mark=0, temp_device="cpu", old_device=None, temp=True, back=True):
    """temp to cpu to calculate and re-back the init device data."""

    @wraps(func)
    def wrapper(*args, **kwargs):

        if temp_device is None:
            device = args[0].device
        else:
            device = torch.device(temp_device) if isinstance(temp_device, str) else temp_device

        if old_device is None:
            device2 = args[mark + 1].device if len(args) > 1 else list(kwargs.values())[0].device
        else:
            device2 = torch.device(old_device) if isinstance(old_device, str) else old_device

        if temp:
            args2 = [args[0]]
            for i in args[1:]:
                try:
                    args2.append(i.to(device=device, copy=False))
                except AttributeError:
                    args2.append(i)
            kwargs2 = {}
            for k, v in kwargs.items():
                try:
                    if isinstance(v, tuple):
                        kwargs2[k] = [i.to(device=device, copy=False) for i in v]
                    else:
                        kwargs2[k] = v.to(device=device, copy=False)
                except AttributeError:
                    kwargs2[k] = v

            result = func(*args2, **kwargs2)
        else:
            result = func(*args, **kwargs)

        if back:
            if isinstance(result, tuple):
                result2 = []
                for i in result:
                    try:
                        result2.append(i.to(device=device2, copy=False))
                    except AttributeError:
                        result2.append(i)
            else:
                try:
                    result2 = result.to(device=device2, copy=False)
                except AttributeError:
                    result2 = result
            return result2
        else:
            return result

    return wrapper


def check_device(mode: Module):
    device = _check_device(mode)
    return torch.device("cpu") if device is None else device


def _check_device(mode: Module):
    device = None
    for i in mode.children():
        if hasattr(i, "weight"):
            device = i.weight.device
            break
        elif hasattr(i, "bias"):
            device = i.bias.device
            break
        elif len(list(i.children())) > 0:
            device = check_device(i)
            if device is not None:
                break
    return device


def get_ptr(index):
    """Trans batch index to ptr"""
    return torch.ops.torch_sparse.ind2ptr(index, index[-1] + 1)


"""This file is not used, which is to fixed one bug in torch_geometrics"""


def separate(cls, batch: BaseData, idx: int, slice_dict: Any,
             inc_dict: Any = None, decrement: bool = True) -> BaseData:
    # Separates the individual element from a `batch` at index `idx`.
    # `separate` can handle both homogeneous and heterogeneous data objects by
    # individually separating all their stores.
    # In addition, `separate` can handle nested data structures such as
    # dictionaries and lists.

    data = cls().stores_as(batch)

    # We iterate over each storage object and recursively separate all its
    # attributes:
    for batch_store, data_store in zip(batch.stores, data.stores):
        key = batch_store._key
        if key is not None:
            attrs = slice_dict[key].keys()
        else:
            attrs = set(batch_store.keys())
            attrs = [attr for attr in slice_dict.keys() if attr in attrs]
        for attr in attrs:
            if key is not None:
                slices = slice_dict[key][attr]
                incs = inc_dict[key][attr] if decrement else None
            else:
                slices = slice_dict[attr]
                incs = inc_dict[attr] if decrement else None
            data_store[attr] = _separate(attr, batch_store[attr], idx, slices,
                                         incs, batch, batch_store, decrement)

        # The `num_nodes` attribute needs special treatment, as we cannot infer
        # the real number of nodes from the total number of nodes alone:
        if hasattr(batch_store, '_num_nodes'):
            data_store.num_nodes = batch_store._num_nodes[idx]

    return data


def _separate(
        key: str,
        value: Any,
        idx: int,
        slices: Any,
        incs: Any,
        batch: BaseData,
        store: BaseStorage,
        decrement: bool,
) -> Any:
    if isinstance(value, Tensor):
        # Narrow a `torch.Tensor` based on `slices`.
        # NOTE: We need to take care of decrementing elements appropriately.
        key = str(key)
        cat_dim = batch.__cat_dim__(key, value, store)
        start, end = int(slices[idx]), int(slices[idx + 1])
        value = value.narrow(cat_dim or 0, start, end - start)
        value = value.squeeze(0) if cat_dim is None else value
        if decrement and (incs.dim() > 1 or int(incs[idx]) != 0):
            value = value - incs[idx].to(value.device)
        return value

    elif isinstance(value, SparseTensor) and decrement:
        # Narrow a `SparseTensor` based on `slices`.
        # NOTE: `cat_dim` may return a tuple to allow for diagonal stacking.
        key = str(key)
        cat_dim = batch.__cat_dim__(key, value, store)
        cat_dims = (cat_dim,) if isinstance(cat_dim, int) else cat_dim
        for i, dim in enumerate(cat_dims):
            start, end = int(slices[idx][i]), int(slices[idx + 1][i])
            value = value.narrow(dim, start, end - start)
        return value

    elif isinstance(value, Mapping):
        # Recursively separate elements of dictionaries.
        return {
            key: _separate(key, elem, idx, slices[key],
                           incs[key] if decrement else None, batch, store,
                           decrement)
            for key, elem in value.items()
        }

    elif (isinstance(value, Sequence) and isinstance(value[0], Sequence)
          and not isinstance(value[0], str) and len(value[0]) > 0
          and isinstance(value[0][0], (Tensor, SparseTensor))
          and isinstance(slices, Sequence)):
        # Recursively separate elements of lists of lists.
        return [elem[idx] for elem in value]

    elif (isinstance(value, Sequence) and not isinstance(value, str)
          and isinstance(value[0], (Tensor, SparseTensor))
          and isinstance(slices, Sequence)):
        # Recursively separate elements of lists of Tensors/SparseTensors.
        return [
            _separate(key, elem, idx, slices[i],
                      incs[i] if decrement else None, batch, store, decrement)
            for i, elem in enumerate(value)
        ]

    else:
        return value[idx]




# Use Asap for a huge performance increase if it is installed
from ase.spacegroup import crystal
from ase.build import cut

# #######221##################0

def perovskite_conv_221(a_atom="Ba", b_atom="Ti", c_atom="O", a=7, size=(1, 1, 1),**kwargs):
    a_atom = [a_atom] * 2 if not isinstance(a_atom, list) else a_atom
    b_atom = [b_atom] * 2 if not isinstance(b_atom, list) else b_atom
    c_atom = [c_atom] * 6 if not isinstance(c_atom, list) else c_atom

    place = [
        [0.5, 0.0, 0.5],
        [0.0, 0.5, 0.5],
        [0.5, 0.5, 0.0],
        [0.0, 0.0, 0.0],

        [0.0, 0.0, 0.5],
        [0.5, 0.5, 0.5],
        [0.25, 0.25, 0.0],
        [0.25, 0.75, 0.0],
        [0.75, 0.25, 0.0],
        [0.75, 0.75, 0.0],

    ]
    ats = a_atom + b_atom + c_atom

    atoms = crystal(ats, place, size=size,
                    cellpar=[a, a, a / 1.4142, 90, 90, 90])
    return atoms


def perovskite_prim_221(a_atom="Ba", b_atom="Ti", c_atom="O", a=7, size=(1, 1, 1),**kwargs):
    a_atom = [a_atom] * 1 if not isinstance(a_atom, list) else a_atom
    b_atom = [b_atom] * 1 if not isinstance(b_atom, list) else b_atom
    c_atom = [c_atom] * 1 if not isinstance(c_atom, list) else c_atom

    place = [
        [0.0, 0.0, 0.0],
        [0.5, 0.5, 0.5],
        [0.0, 0.5, 0.5],
    ]
    ats = a_atom + b_atom + c_atom

    atoms = crystal(ats, place, size=size, spacegroup=221,
                    cellpar=[a, a, a, 90, 90, 90])
    return atoms


# # 轴拉长
# ########127##################
def perovskite_conv_127(a_atom="Ba", b_atom="Ti", c_atom="O", a=7, alpha=1.1, cycle=0.05, size=(1, 1, 1),**kwargs):
    a_atom = [a_atom] * 2 if not isinstance(a_atom, list) else a_atom
    b_atom = [b_atom] * 2 if not isinstance(b_atom, list) else b_atom
    c_atom = [c_atom] * 6 if not isinstance(c_atom, list) else c_atom

    place = [
        [0.5, 0.0, 0.5],
        [0.0, 0.5, 0.5],
        [0.5, 0.5, 0.0],
        [0.0, 0.0, 0.0],

        [0.0, 0.0, 0.5],
        [0.5, 0.5, 0.5],

        [0.25 + cycle, 0.25 - cycle, 0.0],
        [0.25 - cycle, 0.75 - cycle, 0.0],
        [0.75 + cycle, 0.25 + cycle, 0.0],
        [0.75 - cycle, 0.75 + cycle, 0.0],
    ]
    ats = a_atom + b_atom + c_atom

    atoms = crystal(ats, place, size=size,
                    cellpar=[a, a, a / 1.4142 * alpha, 90, 90, 90])
    return atoms


def perovskite_prim_127(a_atom="Ba", b_atom="Ti", c_atom="O", a=7, alpha=1.0, cycle=0.04, size=(1, 1, 1),**kwargs):
    a_atom = [a_atom] * 1 if not isinstance(a_atom, list) else a_atom
    b_atom = [b_atom] * 1 if not isinstance(b_atom, list) else b_atom
    c_atom = [c_atom] * 2 if not isinstance(c_atom, list) else c_atom

    place = [
        [0.0, 0.5, 0],
        [0.0, 0.0, 0.5],
        [0.25 - cycle, 0.25 + cycle, 0.50, ],  # cycle
        [0.0, 0.0, 0.0],  # bond
    ]
    ats = a_atom + b_atom + c_atom

    atoms = crystal(ats, place, size=size, spacegroup=127,
                    cellpar=[a, a, a / 1.4142 * alpha, 90, 90, 90])
    return atoms


# #####62,################################

def perovskite_conv_62(a_atom="Ca", b_atom="Ti", c_atom="O", a=7, alpha=1.0, move=0.04, cycle=0.04, size=(1, 1, 1),**kwargs):
    a_atom = [a_atom] * 1 if not isinstance(a_atom, list) else a_atom
    b_atom = [b_atom] * 1 if not isinstance(b_atom, list) else b_atom
    c_atom = [c_atom] * 2 if not isinstance(c_atom, list) else c_atom

    ats = a_atom + b_atom + c_atom

    place = [
        [0.5 - move, 0.25, 0],
        [0.0, 0.0, 0],
        [0.25 - cycle, 0.0, 0.25 + cycle],  # cycle
        [0.5 + move / 2, 0.25, 0.57 + move],  # bond
    ]

    # ["Ca", "Ti", "O", "O", ],
    atoms = crystal(ats, place, size=(2, 1, 2), spacegroup=62,
                    cellpar=[a, a * 1.4142 * alpha, a, 90, 90, 90], primitive_cell=True)

    atoms2 = cut(atoms, a=(0.5, 0, 0), b=(0, 1, 0), c=(0, 0, 0.5), origo=(0.25, 0, 0.25))
    # change dimension
    position = atoms2.get_positions()
    atoms2.set_positions(position[:, (0, 2, 1)])
    atoms2.set_cell(atoms2.get_cell()[:, (0, 2, 1)])
    atoms2.set_cell(atoms2.get_cell()[(0, 2, 1), :])
    atoms2 *= size

    return atoms2


def perovskite_prim_62(a_atom="Ca", b_atom="Ti", c_atom="O", a=7, alpha=1.0, move=0.04, cycle=0.04, size=(1, 1, 1),**kwargs):
    a_atom = [a_atom] * 1 if not isinstance(a_atom, list) else a_atom
    b_atom = [b_atom] * 1 if not isinstance(b_atom, list) else b_atom
    c_atom = [c_atom] * 2 if not isinstance(c_atom, list) else c_atom

    place = [
        [0.5 - move, 0.25, 0],
        [0.0, 0.0, 0],
        [0.25 - cycle, 0.0, 0.25 + cycle],  # cycle
        [0.5 + move / 2, 0.25, 0.57 + move],  # bond
    ]
    ats = a_atom + b_atom + c_atom
    # ["Ca", "Ti", "O", "O", ],
    atoms = crystal(ats, place, size=size, spacegroup=62,
                    cellpar=[a, a * 1.4142 * alpha, a, 90, 90, 90], primitive_cell=True)
    return atoms


if __name__ == "__main__":
    at = perovskite_prim_221()
    at.rotate(90, (1, 1, 0), rotate_cell=True)
    sg = at.info["spacegroup"]

    prim_cell = sg.scaled_primitive_cell

    # Preserve calculator if present:

    atoms = cut(at, a=prim_cell[0], b=prim_cell[1], c=prim_cell[2])

    from ase.visualize import view

    view(atoms)


import warnings
from typing import List, Optional, Dict

import ase.data as ase_data
import torch
from torch_geometric.data import Data
from torch_geometric.transforms import BaseTransform

from hpc2ml.function.support import add_edge_no_pbc, add_edge_no_pbc_from_index, \
    add_edge_pbc_from_index, add_edge_pbc, \
    distribute_edge


class AddEdgeNoPBC(BaseTransform):
    """Add edge index, with no pbc."""

    def __init__(self, cutoff=5.0):
        self.cutoff = cutoff

    def __call__(self, data):
        if hasattr(data, "edge_index"):
            data = add_edge_no_pbc_from_index(data)
        else:
            data = add_edge_no_pbc(data, cutoff=self.cutoff)
        return data


class AddEdgePBC(BaseTransform):
    """Add edge index, with no pbc."""

    def __init__(self, cutoff=5.0, max_neighbors=16):
        self.cutoff = cutoff
        self.max_neighbors = max_neighbors

    def __call__(self, data):
        if hasattr(data, "edge_index"):
            data = add_edge_pbc_from_index(data)
        else:
            data = add_edge_pbc(data, cutoff=self.cutoff, max_neighbors=self.max_neighbors)
        return data


class AddAttrToWeight(BaseTransform):
    """Add edge index, with no pbc."""

    def __call__(self, data):
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)
        return data


class AddAttrSumToAttrAndWeight(BaseTransform):
    """Add edge index, with no pbc."""

    def __call__(self, data):
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)
            data.edge_attr = data.edge_weight.reshape(-1, 1)
        return data


class AddAttrSumToAttr(BaseTransform):
    """Add edge index, with no pbc."""

    def __call__(self, data):
        assert hasattr(data, "edge_attr") and data.edge_attr is not None
        assert data.edge_attr.shape[1] == 3
        if not hasattr(data, "edge_weight") or data.edge_weight is None:
            data.edge_weight = torch.linalg.norm(data.edge_attr, dim=1)

        wei = data.edge_weight

        data.edge_attr = torch.cat((wei.reshape(-1, 1), data.edge_attr), dim=1)

        return data


class DistributeEdgeAttr(BaseTransform):
    """Compact with rotnet network (deep potential)"""

    def __init__(self, r_cs=2.0, r_c=6.0, cat_weight_attr=True):
        super().__init__()
        self.r_cs = r_cs
        self.r_c = r_c
        self.cat_weight_attr = cat_weight_attr

    def __call__(self, data):
        data = distribute_edge(data, self.r_cs, self.r_c)
        return data


class AtomNumToType(BaseTransform):
    num_types: int
    chemical_symbol_to_type: Optional[Dict[str, int]]
    type_names: List[str]
    _min_Z: int
    from hpc2ml.data import AtomicDataDict

    def __init__(self, *args, type_names: Optional[List[str]] = None,
                 chemical_symbol_to_type: Optional[Dict[str, int]] = None,
                 chemical_symbols: Optional[List[str]] = None, **kwargs):
        super(AtomNumToType, self).__init__(*args, **kwargs)

        if chemical_symbol_to_type is None:
            if chemical_symbols is None:
                chemical_symbols = ase_data.chemical_symbols
            atomic_nums = [ase_data.atomic_numbers[sym] for sym in chemical_symbols]
            chemical_symbols = [e[1] for e in sorted(zip(atomic_nums, chemical_symbols))]
            chemical_symbol_to_type = {k: i for i, k in enumerate(chemical_symbols)}
            del chemical_symbols

        # Build from chem->type mapping, if provided
        self.chemical_symbol_to_type = chemical_symbol_to_type

        # Validate
        for sym, type in self.chemical_symbol_to_type.items():
            assert sym in ase_data.atomic_numbers, f"Invalid chemical symbol {sym}"
            assert 0 <= type, f"Invalid type number {type}"
        assert set(self.chemical_symbol_to_type.values()) == set(
            range(len(self.chemical_symbol_to_type))
        )
        if type_names is None:
            # Make type_names
            type_names = [None] * len(self.chemical_symbol_to_type)
            for sym, type in self.chemical_symbol_to_type.items():
                type_names[type] = sym
        else:
            # Make sure they agree on types
            # We already checked that chem->type is contiguous,
            # so enough to check length since type_names is a list
            assert len(type_names) == len(self.chemical_symbol_to_type)
        # Make mapper array
        valid_atomic_numbers = [
            ase_data.atomic_numbers[sym] for sym in self.chemical_symbol_to_type
        ]
        self._min_Z = min(valid_atomic_numbers)
        self._max_Z = max(valid_atomic_numbers)
        Z_to_index = torch.full(
            size=(1 + self._max_Z - self._min_Z,), fill_value=-1, dtype=torch.long
        )
        for sym, type in self.chemical_symbol_to_type.items():
            Z_to_index[ase_data.atomic_numbers[sym] - self._min_Z] = type
        self._Z_to_index = Z_to_index
        self._index_to_Z = torch.zeros(
            size=(len(self.chemical_symbol_to_type),), dtype=torch.long
        )
        for sym, type_idx in self.chemical_symbol_to_type.items():
            self._index_to_Z[type_idx] = ase_data.atomic_numbers[sym]
        self._valid_set = set(valid_atomic_numbers)

        # check
        if type_names is None:
            raise ValueError(
                "None of chemical_symbols, chemical_symbol_to_type, nor type_names was provided; "
                "exactly one is required")
        # validate type names
        assert all(n.isalnum() for n in type_names), "Type names must contain only alphanumeric characters"
        # Set to however many maps specified -- we already checked contiguous
        self.num_types = len(type_names)
        # Check type_names
        self.type_names = type_names

    def __call__(self, data: Data, types_required: bool = True) -> Data:
        if hasattr(data, self.AtomicDataDict.ATOM_TYPE_KEY):
            if hasattr(data, self.AtomicDataDict.ATOMIC_NUMBERS_KEY):
                warnings.warn(
                    "Data contained both ATOM_TYPE_KEY and ATOMIC_NUMBERS_KEY; ignoring ATOMIC_NUMBERS_KEY"
                )
        elif hasattr(data, self.AtomicDataDict.ATOMIC_NUMBERS_KEY):
            assert (
                    self.chemical_symbol_to_type is not None
            ), "Atomic numbers provided but there is no chemical_symbols/chemical_symbol_to_type mapping!"
            atomic_numbers = getattr(data, self.AtomicDataDict.ATOMIC_NUMBERS_KEY)

            delattr(data, self.AtomicDataDict.ATOMIC_NUMBERS_KEY)

            setattr(data, self.AtomicDataDict.ATOM_TYPE_KEY, self._transform(atomic_numbers))
        else:
            if types_required:
                raise KeyError(
                    "Data doesn't contain any atom type information (ATOM_TYPE_KEY or ATOMIC_NUMBERS_KEY)"
                )
        return data

    def _transform(self, atomic_numbers: torch.Tensor) -> torch.Tensor:
        if atomic_numbers.min() < self._min_Z or atomic_numbers.max() > self._max_Z:
            bad_set = set(torch.unique(atomic_numbers).cpu().tolist()) - self._valid_set
            raise ValueError(
                f"Data included atomic numbers {bad_set} that are not part of the atomic number -> type mapping!"
            )

        return self._Z_to_index.to(device=atomic_numbers.device)[
            atomic_numbers - self._min_Z
            ]
